"""Tests for kicad_tools.drc.suggestions module."""

from pathlib import Path

import pytest

from kicad_tools.drc import (
    DRCReport,
    DRCViolation,
    FixAction,
    FixSuggestion,
    Location,
    Severity,
    ViolationType,
    generate_fix_suggestions,
)
from kicad_tools.drc.suggestions import (
    calculate_clearance_fix,
    direction_name,
)


@pytest.fixture
def sample_drc_report(fixtures_dir: Path) -> Path:
    """Return the path to the sample DRC report."""
    return fixtures_dir / "sample_drc.rpt"


class TestDirectionName:
    """Tests for direction_name helper function."""

    def test_right(self):
        """Test rightward direction."""
        assert direction_name(1.0, 0.0) == "right"

    def test_left(self):
        """Test leftward direction."""
        assert direction_name(-1.0, 0.0) == "left"

    def test_up(self):
        """Test upward direction (negative Y in KiCad)."""
        assert direction_name(0.0, -1.0) == "up"

    def test_down(self):
        """Test downward direction (positive Y in KiCad)."""
        assert direction_name(0.0, 1.0) == "down"

    def test_up_right(self):
        """Test diagonal up-right direction."""
        assert direction_name(1.0, -1.0) == "up-right"

    def test_down_left(self):
        """Test diagonal down-left direction."""
        assert direction_name(-1.0, 1.0) == "down-left"

    def test_in_place(self):
        """Test zero movement."""
        assert direction_name(0.0, 0.0) == "in place"


class TestFixSuggestion:
    """Tests for FixSuggestion dataclass."""

    def test_to_dict(self):
        """Test serialization to dictionary."""
        suggestion = FixSuggestion(
            action=FixAction.MOVE,
            target="R1",
            parameters={"dx": 0.5, "dy": 0.0},
            description="Move R1 0.5mm right",
            priority=1,
            complexity="easy",
        )

        d = suggestion.to_dict()
        assert d["action"] == "move"
        assert d["target"] == "R1"
        assert d["parameters"]["dx"] == 0.5
        assert d["description"] == "Move R1 0.5mm right"

    def test_to_dict_with_alternatives(self):
        """Test serialization includes alternatives."""
        alt = FixSuggestion(
            action=FixAction.REROUTE,
            target="NET1",
            parameters={},
            description="Reroute NET1",
        )
        suggestion = FixSuggestion(
            action=FixAction.MOVE,
            target="R1",
            parameters={},
            description="Move R1",
            alternatives=[alt],
        )

        d = suggestion.to_dict()
        assert len(d["alternatives"]) == 1
        assert d["alternatives"][0]["action"] == "reroute"

    def test_str(self):
        """Test string representation."""
        suggestion = FixSuggestion(
            action=FixAction.MOVE,
            target="R1",
            parameters={},
            description="Move R1 0.5mm right",
        )
        assert str(suggestion) == "Move R1 0.5mm right"


class TestClearanceFix:
    """Tests for clearance violation fix calculation."""

    def test_clearance_fix_with_values(self):
        """Test fix calculation when measurements are available."""
        violation = DRCViolation(
            type=ViolationType.CLEARANCE,
            type_str="clearance",
            severity=Severity.ERROR,
            message="Clearance violation",
            required_value_mm=0.2,
            actual_value_mm=0.15,
            locations=[
                Location(x_mm=100.0, y_mm=50.0, layer="F.Cu"),
                Location(x_mm=100.5, y_mm=50.0, layer="F.Cu"),
            ],
            items=["Pad 1 of R1", "Via [GND]"],
        )

        suggestion = calculate_clearance_fix(violation)
        assert suggestion is not None
        assert suggestion.action == FixAction.MOVE
        assert suggestion.target == "Pad 1 of R1"
        assert "distance_mm" in suggestion.parameters
        # Delta = 0.2 - 0.15 + 0.1 margin = 0.15
        assert suggestion.parameters["distance_mm"] == pytest.approx(0.15, rel=0.01)

    def test_clearance_fix_includes_direction(self):
        """Test that fix includes direction in description."""
        violation = DRCViolation(
            type=ViolationType.CLEARANCE,
            type_str="clearance",
            severity=Severity.ERROR,
            message="Clearance violation",
            required_value_mm=0.2,
            actual_value_mm=0.15,
            locations=[
                Location(x_mm=100.0, y_mm=50.0, layer="F.Cu"),
                Location(x_mm=99.0, y_mm=50.0, layer="F.Cu"),  # Element2 is to the left
            ],
            items=["R1", "U1"],
        )

        suggestion = calculate_clearance_fix(violation)
        assert suggestion is not None
        # Element1 should move away from Element2 (to the right)
        assert "right" in suggestion.description.lower()

    def test_clearance_fix_without_values(self):
        """Test fix when measurements are not available."""
        violation = DRCViolation(
            type=ViolationType.CLEARANCE,
            type_str="clearance",
            severity=Severity.ERROR,
            message="Clearance violation",
        )

        suggestion = calculate_clearance_fix(violation)
        assert suggestion is not None
        assert suggestion.action == FixAction.REROUTE

    def test_clearance_fix_includes_alternatives(self):
        """Test that alternatives are provided."""
        violation = DRCViolation(
            type=ViolationType.CLEARANCE,
            type_str="clearance",
            severity=Severity.ERROR,
            message="Clearance violation",
            required_value_mm=0.2,
            actual_value_mm=0.15,
            locations=[
                Location(x_mm=100.0, y_mm=50.0, layer="F.Cu"),
                Location(x_mm=100.5, y_mm=50.0, layer="F.Cu"),
            ],
            items=["R1", "U1"],
            nets=["VCC"],
        )

        suggestion = calculate_clearance_fix(violation)
        assert suggestion is not None
        assert len(suggestion.alternatives) >= 1
        # Should have alternative to move the other element
        alt_targets = [alt.target for alt in suggestion.alternatives]
        assert "U1" in alt_targets or "VCC" in alt_targets

    def test_non_clearance_returns_none(self):
        """Test that non-clearance violations return None."""
        violation = DRCViolation(
            type=ViolationType.SHORTING_ITEMS,
            type_str="shorting_items",
            severity=Severity.ERROR,
            message="Short circuit",
        )

        suggestion = calculate_clearance_fix(violation)
        assert suggestion is None


class TestGenerateFixSuggestions:
    """Tests for the main generate_fix_suggestions function."""

    def test_clearance_suggestion(self, sample_drc_report: Path):
        """Test generating suggestion for clearance violation."""
        report = DRCReport.load(sample_drc_report)
        clearance_violations = report.by_type(ViolationType.CLEARANCE)

        assert len(clearance_violations) > 0
        suggestion = generate_fix_suggestions(clearance_violations[0])
        assert suggestion is not None
        assert suggestion.action == FixAction.MOVE

    def test_shorting_suggestion(self, sample_drc_report: Path):
        """Test generating suggestion for short circuit."""
        report = DRCReport.load(sample_drc_report)
        shorts = report.by_type(ViolationType.SHORTING_ITEMS)

        assert len(shorts) > 0
        suggestion = generate_fix_suggestions(shorts[0])
        assert suggestion is not None
        assert suggestion.action == FixAction.DELETE

    def test_unconnected_suggestion(self, sample_drc_report: Path):
        """Test generating suggestion for unconnected items."""
        report = DRCReport.load(sample_drc_report)
        unconnected = report.by_type(ViolationType.UNCONNECTED_ITEMS)

        assert len(unconnected) > 0
        suggestion = generate_fix_suggestions(unconnected[0])
        assert suggestion is not None
        assert suggestion.action == FixAction.CONNECT

    def test_track_width_suggestion(self, sample_drc_report: Path):
        """Test generating suggestion for track width violation."""
        report = DRCReport.load(sample_drc_report)
        track_violations = report.by_type(ViolationType.TRACK_WIDTH)

        assert len(track_violations) > 0
        suggestion = generate_fix_suggestions(track_violations[0])
        assert suggestion is not None
        assert suggestion.action == FixAction.RESIZE

    def test_via_annular_width_suggestion(self):
        """Test generating suggestion for via annular width violation."""
        violation = DRCViolation(
            type=ViolationType.VIA_ANNULAR_WIDTH,
            type_str="via_annular_width",
            severity=Severity.ERROR,
            message="Via annular width too small",
            required_value_mm=0.15,
            actual_value_mm=0.10,
        )

        suggestion = generate_fix_suggestions(violation)
        assert suggestion is not None
        assert suggestion.action == FixAction.RESIZE
        assert "pad" in suggestion.description.lower() or "via" in suggestion.description.lower()

    def test_silk_over_copper_suggestion(self):
        """Test generating suggestion for silkscreen over copper."""
        violation = DRCViolation(
            type=ViolationType.SILK_OVER_COPPER,
            type_str="silk_over_copper",
            severity=Severity.WARNING,
            message="Silkscreen over copper",
        )

        suggestion = generate_fix_suggestions(violation)
        assert suggestion is not None
        assert suggestion.action == FixAction.MOVE

    def test_copper_edge_clearance_suggestion(self):
        """Test generating suggestion for copper-to-edge clearance."""
        violation = DRCViolation(
            type=ViolationType.COPPER_EDGE_CLEARANCE,
            type_str="copper_edge_clearance",
            severity=Severity.ERROR,
            message="Copper too close to edge",
            required_value_mm=0.25,
            actual_value_mm=0.15,
        )

        suggestion = generate_fix_suggestions(violation)
        assert suggestion is not None
        assert suggestion.action == FixAction.MOVE
        assert "edge" in suggestion.description.lower()


class TestDRCViolationWithSuggestions:
    """Tests for DRCViolation integration with suggestions."""

    def test_violation_to_dict_includes_suggestions(self):
        """Test that to_dict includes suggestions when populated."""
        violation = DRCViolation(
            type=ViolationType.CLEARANCE,
            type_str="clearance",
            severity=Severity.ERROR,
            message="Clearance violation",
            required_value_mm=0.2,
            actual_value_mm=0.15,
        )

        # Add suggestion
        suggestion = generate_fix_suggestions(violation)
        if suggestion:
            violation.suggestions = [suggestion]

        d = violation.to_dict()
        assert "suggestions" in d
        assert len(d["suggestions"]) == 1
        assert d["suggestions"][0]["action"] == "move" or d["suggestions"][0]["action"] == "reroute"

    def test_violation_delta_mm(self):
        """Test delta_mm property calculation."""
        violation = DRCViolation(
            type=ViolationType.CLEARANCE,
            type_str="clearance",
            severity=Severity.ERROR,
            message="Clearance violation",
            required_value_mm=0.2,
            actual_value_mm=0.15,
        )

        assert violation.delta_mm == pytest.approx(0.05)

    def test_violation_delta_mm_none_when_missing(self):
        """Test delta_mm returns None when values missing."""
        violation = DRCViolation(
            type=ViolationType.CLEARANCE,
            type_str="clearance",
            severity=Severity.ERROR,
            message="Clearance violation",
        )

        assert violation.delta_mm is None

    def test_to_dict_includes_delta(self):
        """Test that to_dict includes delta_mm."""
        violation = DRCViolation(
            type=ViolationType.CLEARANCE,
            type_str="clearance",
            severity=Severity.ERROR,
            message="Clearance violation",
            required_value_mm=0.2,
            actual_value_mm=0.15,
        )

        d = violation.to_dict()
        assert d["delta_mm"] == pytest.approx(0.05)


class TestFixActionEnum:
    """Tests for FixAction enum."""

    def test_all_actions_have_values(self):
        """Test that all actions have string values."""
        for action in FixAction:
            assert isinstance(action.value, str)
            assert len(action.value) > 0

    def test_expected_actions_exist(self):
        """Test that expected action types exist."""
        expected = ["move", "reroute", "resize", "delete", "connect", "adjust_rule"]
        actual = [a.value for a in FixAction]
        for exp in expected:
            assert exp in actual



"""Tests for kicad_tools.validate.consistency module."""

from pathlib import Path

import pytest

from kicad_tools.validate.consistency import (
    ConsistencyIssue,
    ConsistencyResult,
    SchematicPCBChecker,
)


class TestConsistencyIssue:
    """Tests for ConsistencyIssue dataclass."""

    def test_valid_issue_creation(self):
        """Test creating a valid consistency issue."""
        issue = ConsistencyIssue(
            issue_type="missing",
            domain="component",
            schematic_value="R1",
            pcb_value=None,
            reference="R1",
            severity="error",
            suggestion="Add footprint for R1 to PCB",
        )
        assert issue.issue_type == "missing"
        assert issue.domain == "component"
        assert issue.is_error
        assert not issue.is_warning

    def test_warning_issue(self):
        """Test creating a warning issue."""
        issue = ConsistencyIssue(
            issue_type="extra",
            domain="component",
            schematic_value=None,
            pcb_value="TP1",
            reference="TP1",
            severity="warning",
            suggestion="Remove TP1 from PCB or add to schematic",
        )
        assert issue.is_warning
        assert not issue.is_error

    def test_invalid_issue_type_raises(self):
        """Test that invalid issue_type raises ValueError."""
        with pytest.raises(ValueError, match="issue_type must be"):
            ConsistencyIssue(
                issue_type="invalid",
                domain="component",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="error",
                suggestion="Fix it",
            )

    def test_invalid_domain_raises(self):
        """Test that invalid domain raises ValueError."""
        with pytest.raises(ValueError, match="domain must be"):
            ConsistencyIssue(
                issue_type="missing",
                domain="invalid",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="error",
                suggestion="Fix it",
            )

    def test_invalid_severity_raises(self):
        """Test that invalid severity raises ValueError."""
        with pytest.raises(ValueError, match="severity must be"):
            ConsistencyIssue(
                issue_type="missing",
                domain="component",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="critical",
                suggestion="Fix it",
            )

    def test_to_dict(self):
        """Test conversion to dictionary."""
        issue = ConsistencyIssue(
            issue_type="mismatch",
            domain="property",
            schematic_value="10k",
            pcb_value="10K",
            reference="R1",
            severity="warning",
            suggestion="Update R1 value",
        )
        d = issue.to_dict()
        assert d["issue_type"] == "mismatch"
        assert d["domain"] == "property"
        assert d["schematic_value"] == "10k"
        assert d["pcb_value"] == "10K"


class TestConsistencyResult:
    """Tests for ConsistencyResult dataclass."""

    def test_empty_result_is_consistent(self):
        """Test that empty result is consistent."""
        result = ConsistencyResult(issues=[])
        assert result.is_consistent
        assert result.error_count == 0
        assert result.warning_count == 0

    def test_result_with_errors_is_inconsistent(self):
        """Test that result with errors is inconsistent."""
        issues = [
            ConsistencyIssue(
                issue_type="missing",
                domain="component",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="error",
                suggestion="Add R1",
            ),
        ]
        result = ConsistencyResult(issues=issues)
        assert not result.is_consistent
        assert result.error_count == 1

    def test_result_with_only_warnings_is_consistent(self):
        """Test that result with only warnings is still consistent."""
        issues = [
            ConsistencyIssue(
                issue_type="extra",
                domain="component",
                schematic_value=None,
                pcb_value="TP1",
                reference="TP1",
                severity="warning",
                suggestion="Remove TP1",
            ),
        ]
        result = ConsistencyResult(issues=issues)
        assert result.is_consistent
        assert result.warning_count == 1

    def test_filtering_by_domain(self):
        """Test filtering issues by domain."""
        issues = [
            ConsistencyIssue(
                issue_type="missing",
                domain="component",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="error",
                suggestion="Add R1",
            ),
            ConsistencyIssue(
                issue_type="mismatch",
                domain="property",
                schematic_value="10k",
                pcb_value="10K",
                reference="R2",
                severity="warning",
                suggestion="Update value",
            ),
            ConsistencyIssue(
                issue_type="mismatch",
                domain="net",
                schematic_value="VCC",
                pcb_value="VDD",
                reference="U1.1",
                severity="error",
                suggestion="Update net",
            ),
        ]
        result = ConsistencyResult(issues=issues)

        assert len(result.component_issues) == 1
        assert len(result.property_issues) == 1
        assert len(result.net_issues) == 1

    def test_summary(self):
        """Test summary generation."""
        issues = [
            ConsistencyIssue(
                issue_type="missing",
                domain="component",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="error",
                suggestion="Add R1",
            ),
        ]
        result = ConsistencyResult(issues=issues)
        summary = result.summary()
        assert "INCONSISTENT" in summary
        assert "1 errors" in summary


# Fixture: Schematic with component C1 that's not in PCB
SCHEMATIC_WITH_EXTRA_COMPONENT = """(kicad_sch
  (version 20231120)
  (generator "test")
  (generator_version "8.0")
  (uuid "00000000-0000-0000-0000-000000000001")
  (paper "A4")
  (lib_symbols)
  (symbol
    (lib_id "Device:R")
    (at 100 100 0)
    (uuid "00000000-0000-0000-0000-000000000002")
    (property "Reference" "R1" (at 100 90 0) (effects (font (size 1.27 1.27))))
    (property "Value" "10k" (at 100 110 0) (effects (font (size 1.27 1.27))))
    (property "Footprint" "Resistor_SMD:R_0402_1005Metric" (at 100 100 0) (effects (hide yes)))
    (property "Datasheet" "" (at 100 100 0) (effects (hide yes)))
    (pin "1" (uuid "00000000-0000-0000-0000-000000000003"))
    (pin "2" (uuid "00000000-0000-0000-0000-000000000004"))
    (instances
      (project "test"
        (path "/00000000-0000-0000-0000-000000000001"
          (reference "R1")
          (unit 1)
        )
      )
    )
  )
  (symbol
    (lib_id "Device:C")
    (at 120 100 0)
    (uuid "00000000-0000-0000-0000-000000000005")
    (property "Reference" "C1" (at 120 90 0) (effects (font (size 1.27 1.27))))
    (property "Value" "100nF" (at 120 110 0) (effects (font (size 1.27 1.27))))
    (property "Footprint" "Capacitor_SMD:C_0402_1005Metric" (at 120 100 0) (effects (hide yes)))
    (property "Datasheet" "" (at 120 100 0) (effects (hide yes)))
    (pin "1" (uuid "00000000-0000-0000-0000-000000000006"))
    (pin "2" (uuid "00000000-0000-0000-0000-000000000007"))
    (instances
      (project "test"
        (path "/00000000-0000-0000-0000-000000000001"
          (reference "C1")
          (unit 1)
        )
      )
    )
  )
)
"""

# Fixture: PCB with component TP1 that's not in schematic
PCB_WITH_EXTRA_COMPONENT = """(kicad_pcb
  (version 20240108)
  (generator "test")
  (generator_version "8.0")
  (general
    (thickness 1.6)
    (legacy_teardrops no)
  )
  (paper "A4")
  (layers
    (0 "F.Cu" signal)
    (31 "B.Cu" signal)
  )
  (setup
    (pad_to_mask_clearance 0)
  )
  (net 0 "")
  (net 1 "GND")
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "00000000-0000-0000-0000-000000000010")
    (at 100 100)
    (property "Reference" "R1" (at 0 -1.5 0) (layer "F.SilkS") (uuid "00000000-0000-0000-0000-000000000011"))
    (property "Value" "10k" (at 0 1.5 0) (layer "F.Fab") (uuid "00000000-0000-0000-0000-000000000012"))
    (property "Footprint" "Resistor_SMD:R_0402_1005Metric" (at 0 0 0) (layer "F.Fab") (hide yes) (uuid "00000000-0000-0000-0000-000000000013"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64) (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64) (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25))
  )
  (footprint "TestPoint:TestPoint_Pad_1.0x1.0mm"
    (layer "F.Cu")
    (uuid "00000000-0000-0000-0000-000000000020")
    (at 110 100)
    (property "Reference" "TP1" (at 0 -1.5 0) (layer "F.SilkS") (uuid "00000000-0000-0000-0000-000000000021"))
    (property "Value" "TestPoint" (at 0 1.5 0) (layer "F.Fab") (uuid "00000000-0000-0000-0000-000000000022"))
    (pad "1" smd rect (at 0 0) (size 1.0 1.0) (layers "F.Cu" "F.Paste" "F.Mask"))
  )
)
"""

# Fixture: PCB with mismatched value
PCB_WITH_VALUE_MISMATCH = """(kicad_pcb
  (version 20240108)
  (generator "test")
  (generator_version "8.0")
  (general
    (thickness 1.6)
    (legacy_teardrops no)
  )
  (paper "A4")
  (layers
    (0 "F.Cu" signal)
    (31 "B.Cu" signal)
  )
  (setup
    (pad_to_mask_clearance 0)
  )
  (net 0 "")
  (net 1 "GND")
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "00000000-0000-0000-0000-000000000010")
    (at 100 100)
    (property "Reference" "R1" (at 0 -1.5 0) (layer "F.SilkS") (uuid "00000000-0000-0000-0000-000000000011"))
    (property "Value" "4.7k" (at 0 1.5 0) (layer "F.Fab") (uuid "00000000-0000-0000-0000-000000000012"))
    (property "Footprint" "Resistor_SMD:R_0402_1005Metric" (at 0 0 0) (layer "F.Fab") (hide yes) (uuid "00000000-0000-0000-0000-000000000013"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64) (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64) (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25))
  )
)
"""


@pytest.fixture
def schematic_with_extra(tmp_path: Path) -> Path:
    """Create schematic with extra component C1."""
    sch_file = tmp_path / "extra.kicad_sch"
    sch_file.write_text(SCHEMATIC_WITH_EXTRA_COMPONENT)
    return sch_file


@pytest.fixture
def pcb_with_extra(tmp_path: Path) -> Path:
    """Create PCB with extra component TP1."""
    pcb_file = tmp_path / "extra.kicad_pcb"
    pcb_file.write_text(PCB_WITH_EXTRA_COMPONENT)
    return pcb_file


@pytest.fixture
def pcb_with_value_mismatch(tmp_path: Path) -> Path:
    """Create PCB with value mismatch (R1: 4.7k instead of 10k)."""
    pcb_file = tmp_path / "mismatch.kicad_pcb"
    pcb_file.write_text(PCB_WITH_VALUE_MISMATCH)
    return pcb_file


class TestSchematicPCBChecker:
    """Tests for SchematicPCBChecker class."""

    def test_consistent_schematic_pcb(self, minimal_schematic: Path, minimal_pcb: Path):
        """Test that consistent schematic and PCB pass checks."""
        checker = SchematicPCBChecker(minimal_schematic, minimal_pcb)
        result = checker.check()

        # Should be consistent (R1 is in both)
        assert result.is_consistent
        assert result.error_count == 0

    def test_detect_missing_on_pcb(self, schematic_with_extra: Path, pcb_with_extra: Path):
        """Test detection of component in schematic but not on PCB."""
        checker = SchematicPCBChecker(schematic_with_extra, pcb_with_extra)
        result = checker.check()

        # C1 is in schematic but not on PCB
        missing = [i for i in result.component_issues if i.issue_type == "missing"]
        assert len(missing) == 1
        assert missing[0].reference == "C1"
        assert missing[0].is_error

    def test_detect_extra_on_pcb(self, schematic_with_extra: Path, pcb_with_extra: Path):
        """Test detection of component on PCB but not in schematic."""
        checker = SchematicPCBChecker(schematic_with_extra, pcb_with_extra)
        result = checker.check()

        # TP1 is on PCB but not in schematic
        extra = [i for i in result.component_issues if i.issue_type == "extra"]
        assert len(extra) == 1
        assert extra[0].reference == "TP1"
        assert extra[0].is_warning

    def test_detect_value_mismatch(self, minimal_schematic: Path, pcb_with_value_mismatch: Path):
        """Test detection of value property mismatch."""
        checker = SchematicPCBChecker(minimal_schematic, pcb_with_value_mismatch)
        result = checker.check()

        # R1 has value 10k in schematic, 4.7k on PCB
        value_mismatches = [
            i
            for i in result.property_issues
            if i.issue_type == "mismatch" and "value" in i.suggestion.lower()
        ]
        assert len(value_mismatches) == 1
        assert value_mismatches[0].schematic_value == "10k"
        assert value_mismatches[0].pcb_value == "4.7k"

    def test_repr(self, minimal_schematic: Path, minimal_pcb: Path):
        """Test string representation."""
        checker = SchematicPCBChecker(minimal_schematic, minimal_pcb)
        repr_str = repr(checker)
        assert "SchematicPCBChecker" in repr_str
        assert "schematic_symbols=" in repr_str
        assert "pcb_footprints=" in repr_str


class TestConsistencyResultOutput:
    """Tests for ConsistencyResult output methods."""

    def test_to_dict_structure(self):
        """Test to_dict produces correct structure."""
        issues = [
            ConsistencyIssue(
                issue_type="missing",
                domain="component",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="error",
                suggestion="Add R1",
            ),
        ]
        result = ConsistencyResult(issues=issues)
        d = result.to_dict()

        assert "is_consistent" in d
        assert "error_count" in d
        assert "warning_count" in d
        assert "issues" in d
        assert isinstance(d["issues"], list)
        assert len(d["issues"]) == 1

    def test_iteration(self):
        """Test iterating over result."""
        issues = [
            ConsistencyIssue(
                issue_type="missing",
                domain="component",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="error",
                suggestion="Add R1",
            ),
            ConsistencyIssue(
                issue_type="extra",
                domain="component",
                schematic_value=None,
                pcb_value="TP1",
                reference="TP1",
                severity="warning",
                suggestion="Remove TP1",
            ),
        ]
        result = ConsistencyResult(issues=issues)

        refs = [i.reference for i in result]
        assert refs == ["R1", "TP1"]

    def test_len(self):
        """Test len() on result."""
        issues = [
            ConsistencyIssue(
                issue_type="missing",
                domain="component",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="error",
                suggestion="Add R1",
            ),
        ]
        result = ConsistencyResult(issues=issues)
        assert len(result) == 1

    def test_bool_empty(self):
        """Test bool() on empty result."""
        result = ConsistencyResult(issues=[])
        assert not result

    def test_bool_with_issues(self):
        """Test bool() on result with issues."""
        issues = [
            ConsistencyIssue(
                issue_type="missing",
                domain="component",
                schematic_value="R1",
                pcb_value=None,
                reference="R1",
                severity="error",
                suggestion="Add R1",
            ),
        ]
        result = ConsistencyResult(issues=issues)
        assert result



"""TraceOptimizer class for PCB trace cleanup and simplification."""

from __future__ import annotations

from ..layers import Layer
from ..primitives import Route, Segment
from .algorithms import (
    _find_staircase_end,
    _optimal_path,
    compress_staircase,
    convert_corners_45,
    eliminate_zigzags,
    merge_collinear,
)
from .chain import sort_into_chains
from .collision import CollisionChecker
from .config import OptimizationConfig, OptimizationStats
from .geometry import (
    angle_between,
    count_corners,
    is_90_degree_corner,
    is_connected,
    is_zigzag,
    same_direction,
    segment_direction,
    segments_touch,
    shorten_segment_end,
    shorten_segment_start,
    total_length,
)
from .pcb import optimize_pcb, parse_net_names, parse_segments, replace_segments


class TraceOptimizer:
    """Optimizer for PCB trace cleanup and simplification.

    Optionally uses a collision checker to ensure optimizations don't
    create DRC violations (shorts, track crossings with other nets).
    When a collision checker is provided, optimizations that would create
    collisions are skipped, preserving the original path.

    Example::

        from kicad_tools.router import TraceOptimizer, OptimizationConfig

        # Optimize a route in memory (no collision checking)
        optimizer = TraceOptimizer()
        optimized_route = optimizer.optimize_route(route)

        # Optimize with collision checking
        from kicad_tools.router import GridCollisionChecker
        checker = GridCollisionChecker(grid)
        optimizer = TraceOptimizer(collision_checker=checker)
        optimized_route = optimizer.optimize_route(route)

        # Optimize traces in a PCB file
        stats = optimizer.optimize_pcb("board.kicad_pcb", output="optimized.kicad_pcb")
        print(f"Reduced segments from {stats['before']} to {stats['after']}")
    """

    def __init__(
        self,
        config: OptimizationConfig | None = None,
        collision_checker: CollisionChecker | None = None,
    ):
        """Initialize the trace optimizer.

        Args:
            config: Optimization configuration. Uses defaults if None.
            collision_checker: Optional collision checker for DRC-safe optimization.
                When provided, optimizations that would create collisions are skipped.
                When None, no collision checking is performed (original behavior).
        """
        self.config = config or OptimizationConfig()
        self.collision_checker = collision_checker

    def optimize_segments(self, segments: list[Segment]) -> list[Segment]:
        """Optimize a list of segments for a single net/layer.

        Applies enabled optimizations in order:
        1. Sort segments into connected chains (to avoid cross-chain shortcuts)
        2. Collinear segment merging
        3. Zigzag elimination
        4. Staircase compression
        5. 45-degree corner conversion

        Args:
            segments: List of segments to optimize (may contain multiple chains).

        Returns:
            Optimized list of segments.
        """
        if not segments:
            return []

        # Sort segments into connected chains to prevent cross-chain shortcuts
        chains = self._sort_into_chains(segments)

        # Optimize each chain independently
        all_optimized: list[Segment] = []
        for chain in chains:
            result = list(chain)

            # Apply optimizations in order
            if self.config.merge_collinear:
                result = self.merge_collinear(result)

            if self.config.eliminate_zigzags:
                result = self.eliminate_zigzags(result)

            if self.config.compress_staircase:
                result = self.compress_staircase(result)

            if self.config.convert_45_corners:
                result = self.convert_corners_45(result)

            all_optimized.extend(result)

        return all_optimized

    def _path_is_clear(self, seg: Segment) -> bool:
        """Check if a segment's path is clear using the collision checker.

        Args:
            seg: The segment to check.

        Returns:
            True if path is clear (or no collision checker), False if blocked.
        """
        if self.collision_checker is None:
            return True  # No collision checking - allow all paths

        return self.collision_checker.path_is_clear(
            x1=seg.x1,
            y1=seg.y1,
            x2=seg.x2,
            y2=seg.y2,
            layer=seg.layer,
            width=seg.width,
            exclude_net=seg.net,
        )

    def merge_collinear(self, segments: list[Segment]) -> list[Segment]:
        """Merge adjacent collinear segments.

        Combines segments that:
        - Are connected (end of one matches start of next)
        - Have the same direction
        - Are on the same layer
        - Would not cross obstacles (if collision checker provided)

        Args:
            segments: List of segments to merge.

        Returns:
            List with collinear segments merged.
        """
        return merge_collinear(segments, self.config, self._path_is_clear)

    def eliminate_zigzags(self, segments: list[Segment]) -> list[Segment]:
        """Remove unnecessary zigzag patterns.

        Identifies segments where the path backtracks and removes
        the unnecessary detour, but only if the shortcut path is clear.

        Args:
            segments: List of segments to process.

        Returns:
            List with zigzags eliminated.
        """
        return eliminate_zigzags(segments, self.config, self._path_is_clear)

    def compress_staircase(self, segments: list[Segment]) -> list[Segment]:
        """Compress staircase patterns into optimal diagonal+orthogonal paths.

        Identifies runs of segments alternating between two directions
        (e.g., horizontal and 45° diagonal) and replaces them with an
        optimal 2-3 segment path, but only if the replacement is clear.

        Args:
            segments: List of segments to process.

        Returns:
            List with staircase patterns compressed.
        """
        return compress_staircase(segments, self.config, self._path_is_clear)

    def convert_corners_45(self, segments: list[Segment]) -> list[Segment]:
        """Convert 90-degree corners to 45-degree chamfers.

        Replaces sharp 90-degree turns with smoother 45-degree entry/exit,
        but only if the chamfer path is clear of obstacles.

        Args:
            segments: List of segments to process.

        Returns:
            List with corners converted to 45 degrees.
        """
        return convert_corners_45(segments, self.config, self._path_is_clear)

    def optimize_route(self, route: Route) -> Route:
        """Optimize a complete route.

        Segments are grouped by layer and then sorted into connected chains
        before optimization. This prevents optimization from creating
        shortcuts between unconnected parts of the route.

        Args:
            route: Route to optimize.

        Returns:
            New Route with optimized segments.
        """
        # Group segments by layer for optimization
        segments_by_layer: dict[Layer, list[Segment]] = {}
        for seg in route.segments:
            if seg.layer not in segments_by_layer:
                segments_by_layer[seg.layer] = []
            segments_by_layer[seg.layer].append(seg)

        # Optimize each layer's segments (chain sorting happens in optimize_segments)
        optimized_segments: list[Segment] = []
        for _layer, segs in segments_by_layer.items():
            optimized = self.optimize_segments(segs)
            optimized_segments.extend(optimized)

        return Route(
            net=route.net,
            net_name=route.net_name,
            segments=optimized_segments,
            vias=list(route.vias),  # Vias unchanged
        )

    def optimize_pcb(
        self,
        pcb_path: str,
        output_path: str | None = None,
        net_filter: str | None = None,
        dry_run: bool = False,
    ) -> OptimizationStats:
        """Optimize traces in a PCB file.

        Args:
            pcb_path: Path to input .kicad_pcb file.
            output_path: Path for output file. If None, modifies in place.
            net_filter: Only optimize nets matching this pattern.
            dry_run: If True, calculate stats but don't write output.

        Returns:
            Statistics about the optimization.
        """
        return optimize_pcb(
            pcb_path=pcb_path,
            output_path=output_path,
            optimize_fn=self.optimize_segments,
            config=self.config,
            net_filter=net_filter,
            dry_run=dry_run,
        )

    # =========================================================================
    # Helper methods exposed for tests
    # =========================================================================

    def _is_connected(self, s1: Segment, s2: Segment) -> bool:
        """Check if end of s1 connects to start of s2."""
        return is_connected(s1, s2, self.config.tolerance)

    def _segments_touch(self, s1: Segment, s2: Segment) -> bool:
        """Check if two segments share any endpoint (regardless of direction)."""
        return segments_touch(s1, s2, self.config.tolerance)

    def _sort_into_chains(self, segments: list[Segment]) -> list[list[Segment]]:
        """Sort segments into connected chains."""
        return sort_into_chains(segments, self.config.tolerance)

    def _same_direction(self, s1: Segment, s2: Segment) -> bool:
        """Check if two segments have the same direction."""
        return same_direction(s1, s2, self.config.tolerance)

    def _is_zigzag(self, s1: Segment, s2: Segment, s3: Segment) -> bool:
        """Check if s2 is a zigzag (backtrack) between s1 and s3."""
        return is_zigzag(s1, s2, s3, self.config.tolerance)

    def _angle_between(self, s1: Segment, s2: Segment) -> float:
        """Calculate angle between two segments in degrees (0-180)."""
        return angle_between(s1, s2, self.config.tolerance)

    def _is_90_degree_corner(self, s1: Segment, s2: Segment) -> bool:
        """Check if two segments form a 90-degree corner."""
        return is_90_degree_corner(s1, s2)

    def _shorten_segment_end(self, seg: Segment, amount: float) -> Segment | None:
        """Shorten a segment from its end by the given amount."""
        return shorten_segment_end(seg, amount, self.config.min_segment_length)

    def _shorten_segment_start(self, seg: Segment, amount: float) -> Segment | None:
        """Shorten a segment from its start by the given amount."""
        return shorten_segment_start(seg, amount, self.config.min_segment_length)

    def _count_corners(self, segments: list[Segment]) -> int:
        """Count number of corners (direction changes) in a segment list."""
        return count_corners(segments, self.config.tolerance)

    def _total_length(self, segments: list[Segment]) -> float:
        """Calculate total length of segments."""
        return total_length(segments)

    def _segment_direction(self, seg: Segment) -> float:
        """Calculate the direction of a segment in degrees (0-360)."""
        return segment_direction(seg, self.config.tolerance)

    def _find_staircase_end(self, segments: list[Segment], start_idx: int) -> int:
        """Find the end index of a staircase pattern starting at start_idx."""
        return _find_staircase_end(segments, start_idx, self.config)

    def _optimal_path(
        self,
        start: tuple[float, float],
        end: tuple[float, float],
        template: Segment,
    ) -> list[Segment]:
        """Generate an optimal 2-3 segment path from start to end."""
        return _optimal_path(start, end, template, self.config)

    def _parse_net_names(self, pcb_text: str) -> dict[int, str]:
        """Parse net ID to name mapping from PCB file."""
        return parse_net_names(pcb_text)

    def _parse_segments(self, pcb_text: str) -> dict[str, list[Segment]]:
        """Parse segments from PCB file text, grouped by net name."""
        return parse_segments(pcb_text)

    def _replace_segments(
        self,
        pcb_text: str,
        original: dict[str, list[Segment]],
        optimized: dict[str, list[Segment]],
    ) -> str:
        """Replace original segments with optimized ones in PCB text."""
        return replace_segments(pcb_text, original, optimized)



"""Tests for net connectivity status analysis and CLI command."""

import json
from pathlib import Path

import pytest

from kicad_tools.analysis.net_status import (
    NetStatus,
    NetStatusAnalyzer,
    NetStatusResult,
    PadInfo,
)
from kicad_tools.cli.net_status_cmd import main as net_status_main

# PCB with fully routed nets
FULLY_ROUTED_PCB = """(kicad_pcb
  (version 20240108)
  (generator "test")
  (generator_version "8.0")
  (general (thickness 1.6) (legacy_teardrops no))
  (paper "A4")
  (layers
    (0 "F.Cu" signal)
    (31 "B.Cu" signal)
    (44 "Edge.Cuts" user)
  )
  (setup (pad_to_mask_clearance 0))
  (net 0 "")
  (net 1 "GND")
  (net 2 "+3.3V")
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "fp1")
    (at 100 100)
    (property "Reference" "R1" (at 0 -1.5 0) (layer "F.SilkS") (uuid "ref1"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 2 "+3.3V"))
  )
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "fp2")
    (at 110 100)
    (property "Reference" "R2" (at 0 -1.5 0) (layer "F.SilkS") (uuid "ref2"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 2 "+3.3V"))
  )
  (segment (start 99.49 100) (end 109.49 100) (width 0.25) (layer "F.Cu") (net 1) (uuid "seg1"))
  (segment (start 100.51 100) (end 110.51 100) (width 0.25) (layer "F.Cu") (net 2) (uuid "seg2"))
)
"""


# PCB with incomplete routing (GND net only has one segment)
INCOMPLETE_PCB = """(kicad_pcb
  (version 20240108)
  (generator "test")
  (generator_version "8.0")
  (general (thickness 1.6) (legacy_teardrops no))
  (paper "A4")
  (layers
    (0 "F.Cu" signal)
    (31 "B.Cu" signal)
    (44 "Edge.Cuts" user)
  )
  (setup (pad_to_mask_clearance 0))
  (net 0 "")
  (net 1 "GND")
  (net 2 "+3.3V")
  (net 3 "UNROUTED")
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "fp1")
    (at 100 100)
    (property "Reference" "R1" (at 0 -1.5 0) (layer "F.SilkS") (uuid "ref1"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 2 "+3.3V"))
  )
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "fp2")
    (at 110 100)
    (property "Reference" "R2" (at 0 -1.5 0) (layer "F.SilkS") (uuid "ref2"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 2 "+3.3V"))
  )
  (footprint "Capacitor_SMD:C_0402_1005Metric"
    (layer "F.Cu")
    (uuid "fp3")
    (at 120 100)
    (property "Reference" "C1" (at 0 -1.5 0) (layer "F.SilkS") (uuid "ref3"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 3 "UNROUTED"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 3 "UNROUTED"))
  )
  (segment (start 99.49 100) (end 109.49 100) (width 0.25) (layer "F.Cu") (net 1) (uuid "seg1"))
)
"""


# PCB with via-to-zone connectivity (Issue #419)
# Tests that pads connected to zones via vias are recognized as connected
VIA_ZONE_CONNECTIVITY_PCB = """(kicad_pcb
  (version 20240108)
  (generator "test")
  (generator_version "8.0")
  (general (thickness 1.6) (legacy_teardrops no))
  (paper "A4")
  (layers
    (0 "F.Cu" signal)
    (1 "In1.Cu" signal)
    (2 "In2.Cu" signal)
    (31 "B.Cu" signal)
    (44 "Edge.Cuts" user)
  )
  (setup (pad_to_mask_clearance 0))
  (net 0 "")
  (net 1 "GND")
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "fp1")
    (at 100 100)
    (property "Reference" "R1" (at 0 -1.5 0) (layer "F.SilkS") (uuid "ref1"))
    (pad "1" smd roundrect (at 0 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
  )
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "fp2")
    (at 150 100)
    (property "Reference" "R2" (at 0 -1.5 0) (layer "F.SilkS") (uuid "ref2"))
    (pad "1" smd roundrect (at 0 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
  )
  (segment (start 100 100) (end 105 100) (width 0.25) (layer "F.Cu") (net 1) (uuid "seg1"))
  (segment (start 150 100) (end 145 100) (width 0.25) (layer "F.Cu") (net 1) (uuid "seg2"))
  (via (at 105 100) (size 0.8) (drill 0.4) (layers "F.Cu" "In1.Cu") (net 1) (uuid "via1"))
  (via (at 145 100) (size 0.8) (drill 0.4) (layers "F.Cu" "In1.Cu") (net 1) (uuid "via2"))
  (zone (net 1) (net_name "GND") (layer "In1.Cu")
    (uuid "zone1")
    (connect_pads (clearance 0.3))
    (min_thickness 0.2)
    (fill yes (thermal_gap 0.3) (thermal_bridge_width 0.3))
    (polygon (pts (xy 90 90) (xy 160 90) (xy 160 110) (xy 90 110)))
    (filled_polygon (layer "In1.Cu") (pts (xy 90 90) (xy 160 90) (xy 160 110) (xy 90 110)))
  )
)
"""


# PCB with zone (plane net)
ZONE_PCB = """(kicad_pcb
  (version 20240108)
  (generator "test")
  (generator_version "8.0")
  (general (thickness 1.6) (legacy_teardrops no))
  (paper "A4")
  (layers
    (0 "F.Cu" signal)
    (31 "B.Cu" signal)
    (44 "Edge.Cuts" user)
  )
  (setup (pad_to_mask_clearance 0))
  (net 0 "")
  (net 1 "GND")
  (net 2 "+3.3V")
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "fp1")
    (at 110 110)
    (property "Reference" "R1" (at 0 -1.5 0) (layer "F.SilkS") (uuid "ref1"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 2 "+3.3V"))
  )
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (uuid "fp2")
    (at 120 110)
    (property "Reference" "R2" (at 0 -1.5 0) (layer "F.SilkS") (uuid "ref2"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 1 "GND"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64)
      (layers "F.Cu" "F.Paste" "F.Mask") (roundrect_rratio 0.25) (net 2 "+3.3V"))
  )
  (zone (net 1) (net_name "GND") (layer "B.Cu")
    (uuid "zone1")
    (connect_pads (clearance 0.3))
    (min_thickness 0.2)
    (fill yes (thermal_gap 0.3) (thermal_bridge_width 0.3))
    (polygon (pts (xy 100 100) (xy 140 100) (xy 140 130) (xy 100 130)))
    (filled_polygon (layer "B.Cu") (pts (xy 100 100) (xy 140 100) (xy 140 130) (xy 100 130)))
  )
)
"""


@pytest.fixture
def fully_routed_pcb(tmp_path: Path) -> Path:
    """Create a fully routed PCB file for testing."""
    pcb_file = tmp_path / "fully_routed.kicad_pcb"
    pcb_file.write_text(FULLY_ROUTED_PCB)
    return pcb_file


@pytest.fixture
def incomplete_pcb(tmp_path: Path) -> Path:
    """Create an incomplete PCB file for testing."""
    pcb_file = tmp_path / "incomplete.kicad_pcb"
    pcb_file.write_text(INCOMPLETE_PCB)
    return pcb_file


@pytest.fixture
def zone_pcb(tmp_path: Path) -> Path:
    """Create a PCB with zones for testing."""
    pcb_file = tmp_path / "zone.kicad_pcb"
    pcb_file.write_text(ZONE_PCB)
    return pcb_file


@pytest.fixture
def via_zone_connectivity_pcb(tmp_path: Path) -> Path:
    """Create a PCB with via-to-zone connectivity for testing (Issue #419)."""
    pcb_file = tmp_path / "via_zone.kicad_pcb"
    pcb_file.write_text(VIA_ZONE_CONNECTIVITY_PCB)
    return pcb_file


class TestPadInfo:
    """Tests for PadInfo dataclass."""

    def test_full_name(self):
        """Test full_name property."""
        pad = PadInfo(
            reference="R1",
            pad_number="2",
            position=(100.0, 100.0),
            is_connected=True,
        )
        assert pad.full_name == "R1.2"

    def test_position(self):
        """Test position storage."""
        pad = PadInfo(
            reference="C1",
            pad_number="1",
            position=(50.5, 75.25),
            is_connected=False,
        )
        assert pad.position == (50.5, 75.25)


class TestNetStatus:
    """Tests for NetStatus dataclass."""

    def test_complete_net(self):
        """Test a fully connected net."""
        net = NetStatus(
            net_number=1,
            net_name="GND",
            total_pads=4,
            connected_pads=[
                PadInfo("R1", "1", (100, 100), True),
                PadInfo("R2", "1", (110, 100), True),
                PadInfo("C1", "1", (120, 100), True),
                PadInfo("C2", "1", (130, 100), True),
            ],
            unconnected_pads=[],
        )
        assert net.status == "complete"
        assert net.connected_count == 4
        assert net.unconnected_count == 0
        assert net.connection_percentage == 100.0

    def test_incomplete_net(self):
        """Test a partially connected net."""
        net = NetStatus(
            net_number=2,
            net_name="+3.3V",
            total_pads=3,
            connected_pads=[
                PadInfo("R1", "2", (100.51, 100), True),
                PadInfo("R2", "2", (110.51, 100), True),
            ],
            unconnected_pads=[
                PadInfo("C1", "1", (120.51, 100), False),
            ],
        )
        assert net.status == "incomplete"
        assert net.connected_count == 2
        assert net.unconnected_count == 1
        assert net.connection_percentage == pytest.approx(66.67, rel=0.01)

    def test_unrouted_net(self):
        """Test a net with no routing."""
        net = NetStatus(
            net_number=3,
            net_name="SIGNAL",
            total_pads=2,
            connected_pads=[],
            unconnected_pads=[
                PadInfo("U1", "5", (100, 100), False),
                PadInfo("U2", "3", (150, 100), False),
            ],
        )
        assert net.status == "unrouted"
        assert net.connected_count == 0
        assert net.unconnected_count == 2

    def test_single_pad_net_complete(self):
        """Test a single-pad net is always complete."""
        net = NetStatus(
            net_number=4,
            net_name="NC",
            total_pads=1,
            connected_pads=[
                PadInfo("U1", "NC", (100, 100), True),
            ],
            unconnected_pads=[],
        )
        assert net.status == "complete"

    def test_plane_net(self):
        """Test plane net identification."""
        net = NetStatus(
            net_number=1,
            net_name="GND",
            is_plane_net=True,
            plane_layer="B.Cu",
        )
        assert net.net_type == "plane"
        assert net.plane_layer == "B.Cu"

    def test_power_net_identification(self):
        """Test power net identification from name patterns."""
        for name in ["+3.3V", "-5V", "VCC", "VDD", "GND", "AGND"]:
            net = NetStatus(net_number=1, net_name=name)
            assert net.net_type == "power", f"{name} should be identified as power"

    def test_signal_net_identification(self):
        """Test signal net identification."""
        net = NetStatus(net_number=1, net_name="SPI_CLK")
        assert net.net_type == "signal"

    def test_to_dict(self):
        """Test serialization to dictionary."""
        net = NetStatus(
            net_number=1,
            net_name="GND",
            total_pads=2,
            connected_pads=[PadInfo("R1", "1", (100, 100), True)],
            unconnected_pads=[PadInfo("R2", "1", (110, 100), False)],
            is_plane_net=True,
            plane_layer="B.Cu",
        )
        d = net.to_dict()
        assert d["net_name"] == "GND"
        assert d["status"] == "incomplete"
        assert d["is_plane_net"] is True
        assert d["plane_layer"] == "B.Cu"
        assert len(d["connected_pads"]) == 1
        assert len(d["unconnected_pads"]) == 1


class TestNetStatusResult:
    """Tests for NetStatusResult dataclass."""

    def test_counts(self):
        """Test result count properties."""
        result = NetStatusResult(
            nets=[
                NetStatus(
                    1,
                    "GND",
                    total_pads=2,
                    connected_pads=[
                        PadInfo("R1", "1", (100, 100), True),
                        PadInfo("R2", "1", (110, 100), True),
                    ],
                    unconnected_pads=[],
                ),
                NetStatus(
                    2,
                    "+3.3V",
                    total_pads=2,
                    connected_pads=[
                        PadInfo("R1", "2", (100.51, 100), True),
                    ],
                    unconnected_pads=[
                        PadInfo("R2", "2", (110.51, 100), False),
                    ],
                ),
                NetStatus(
                    3,
                    "SIG",
                    total_pads=2,
                    connected_pads=[],
                    unconnected_pads=[
                        PadInfo("U1", "1", (100, 100), False),
                        PadInfo("U2", "1", (110, 100), False),
                    ],
                ),
            ],
            total_nets=3,
        )
        assert result.complete_count == 1
        assert result.incomplete_count == 1
        assert result.unrouted_count == 1
        assert result.total_unconnected_pads == 3

    def test_get_net(self):
        """Test getting a specific net by name."""
        result = NetStatusResult(
            nets=[
                NetStatus(1, "GND"),
                NetStatus(2, "+3.3V"),
            ],
            total_nets=2,
        )
        gnd = result.get_net("GND")
        assert gnd is not None
        assert gnd.net_name == "GND"
        assert result.get_net("NONEXISTENT") is None

    def test_summary(self):
        """Test summary generation."""
        result = NetStatusResult(total_nets=5)
        summary = result.summary()
        assert "5 nets" in summary


class TestNetStatusAnalyzer:
    """Tests for NetStatusAnalyzer class."""

    def test_analyze_fully_routed(self, fully_routed_pcb: Path):
        """Test analysis of fully routed PCB."""
        analyzer = NetStatusAnalyzer(fully_routed_pcb)
        result = analyzer.analyze()

        assert result.total_nets == 2  # GND and +3.3V
        assert result.incomplete_count == 0
        assert result.unrouted_count == 0
        assert result.complete_count == 2

    def test_analyze_incomplete(self, incomplete_pcb: Path):
        """Test analysis of PCB with incomplete routing."""
        analyzer = NetStatusAnalyzer(incomplete_pcb)
        result = analyzer.analyze()

        assert result.total_nets == 3  # GND, +3.3V, UNROUTED

        # GND is connected via segment
        gnd = result.get_net("GND")
        assert gnd is not None
        assert gnd.status == "complete"

        # +3.3V has no routing, should be unrouted or incomplete
        v33 = result.get_net("+3.3V")
        assert v33 is not None
        assert v33.status in ("unrouted", "incomplete")  # No segments for this net

        # UNROUTED net has no routing, may be classified as unrouted or incomplete
        unrouted = result.get_net("UNROUTED")
        assert unrouted is not None
        assert unrouted.status in ("unrouted", "incomplete")  # No segments for this net

    def test_analyze_zone_pcb(self, zone_pcb: Path):
        """Test analysis identifies plane nets."""
        analyzer = NetStatusAnalyzer(zone_pcb)
        result = analyzer.analyze()

        gnd = result.get_net("GND")
        assert gnd is not None
        assert gnd.is_plane_net is True
        assert gnd.plane_layer == "B.Cu"

    def test_unconnected_pads_have_positions(self, incomplete_pcb: Path):
        """Test that unconnected pads include position information."""
        analyzer = NetStatusAnalyzer(incomplete_pcb)
        result = analyzer.analyze()

        v33 = result.get_net("+3.3V")
        assert v33 is not None
        assert len(v33.unconnected_pads) > 0

        for pad in v33.unconnected_pads:
            assert pad.position is not None
            assert isinstance(pad.position, tuple)
            assert len(pad.position) == 2

    def test_via_zone_connectivity(self, via_zone_connectivity_pcb: Path):
        """Test that pads connected to zones via vias are recognized as connected.

        Issue #419: net-status doesn't detect via connectivity to planes.

        This test creates a PCB with:
        - Two pads on F.Cu (R1.1 at 100,100 and R2.1 at 150,100)
        - Traces from each pad to vias
        - Vias spanning F.Cu to In1.Cu
        - A GND zone on In1.Cu covering both via positions

        Both pads should be recognized as connected because they both connect
        to the zone via their respective vias (stitching vias pattern).
        """
        analyzer = NetStatusAnalyzer(via_zone_connectivity_pcb)
        result = analyzer.analyze()

        gnd = result.get_net("GND")
        assert gnd is not None, "GND net should exist"
        assert gnd.is_plane_net is True, "GND should be identified as plane net"
        assert gnd.status == "complete", (
            f"GND should be complete (pads connected via zone), "
            f"but found {gnd.unconnected_count} unconnected pads: "
            f"{[p.full_name for p in gnd.unconnected_pads]}"
        )
        assert gnd.total_pads == 2, "GND should have 2 pads"
        assert gnd.connected_count == 2, "Both pads should be connected via zone"
        assert gnd.unconnected_count == 0, "No pads should be unconnected"


class TestNetStatusCLI:
    """Tests for net-status CLI command."""

    def test_basic_usage(self, fully_routed_pcb: Path):
        """Test basic command execution."""
        exit_code = net_status_main([str(fully_routed_pcb)])
        assert exit_code == 0  # All complete

    def test_exit_code_incomplete(self, incomplete_pcb: Path):
        """Test exit code when incomplete nets exist."""
        exit_code = net_status_main([str(incomplete_pcb)])
        assert exit_code == 2  # Incomplete nets found

    def test_json_format(self, fully_routed_pcb: Path, capsys):
        """Test JSON output format."""
        exit_code = net_status_main([str(fully_routed_pcb), "--format", "json"])
        assert exit_code == 0

        captured = capsys.readouterr()
        data = json.loads(captured.out)
        assert "summary" in data
        assert "nets" in data
        assert data["summary"]["total_nets"] == 2

    def test_incomplete_filter(self, incomplete_pcb: Path, capsys):
        """Test --incomplete filter."""
        exit_code = net_status_main([str(incomplete_pcb), "--incomplete"])
        assert exit_code == 2

        captured = capsys.readouterr()
        # Should only show incomplete/unrouted nets
        assert "+3.3V" in captured.out or "UNROUTED" in captured.out

    def test_specific_net(self, incomplete_pcb: Path, capsys):
        """Test --net filter for specific net."""
        _exit_code = net_status_main([str(incomplete_pcb), "--net", "GND"])

        captured = capsys.readouterr()
        assert "GND" in captured.out

    def test_nonexistent_net(self, fully_routed_pcb: Path):
        """Test error when net doesn't exist."""
        exit_code = net_status_main([str(fully_routed_pcb), "--net", "DOES_NOT_EXIST"])
        assert exit_code == 1  # Error

    def test_nonexistent_file(self, tmp_path: Path):
        """Test error when file doesn't exist."""
        exit_code = net_status_main([str(tmp_path / "nonexistent.kicad_pcb")])
        assert exit_code == 1

    def test_verbose_mode(self, incomplete_pcb: Path, capsys):
        """Test verbose output shows all pads."""
        _exit_code = net_status_main([str(incomplete_pcb), "--verbose"])

        captured = capsys.readouterr()
        # Verbose should show coordinate details
        assert "@" in captured.out or "(" in captured.out

    def test_by_class_grouping(self, fully_routed_pcb: Path, capsys):
        """Test --by-class grouping option."""
        exit_code = net_status_main([str(fully_routed_pcb), "--by-class"])
        assert exit_code == 0

        captured = capsys.readouterr()
        assert "Net Class:" in captured.out


class TestNetStatusIntegration:
    """Integration tests for net status functionality."""

    def test_unified_cli_access(self, fully_routed_pcb: Path):
        """Test access via unified kicad-tools CLI."""
        from kicad_tools.cli import main

        exit_code = main(["net-status", str(fully_routed_pcb)])
        assert exit_code == 0

    def test_json_serialization_roundtrip(self, incomplete_pcb: Path):
        """Test that JSON output can be parsed back."""
        analyzer = NetStatusAnalyzer(incomplete_pcb)
        result = analyzer.analyze()

        # Serialize to JSON
        data = result.to_dict()
        json_str = json.dumps(data)

        # Parse back
        parsed = json.loads(json_str)
        assert parsed["total_nets"] == result.total_nets
        assert len(parsed["nets"]) == len(result.nets)



"""
Netlist analysis and comparison CLI commands.

Provides CLI commands for analyzing netlist connectivity, finding issues,
and comparing netlists between schematic versions.

Usage:
    kct netlist analyze design.kicad_sch
    kct netlist list design.kicad_sch --format table
    kct netlist show design.kicad_sch --net VCC
    kct netlist check design.kicad_sch
    kct netlist compare old.kicad_sch new.kicad_sch
    kct netlist export design.kicad_sch -o output.net
"""

from __future__ import annotations

import argparse
import json
import sys
from pathlib import Path

from kicad_tools.operations.netlist import (
    Netlist,
    NetlistNet,
    export_netlist,
)


def main(argv: list[str] | None = None) -> int:
    """Main entry point for kct netlist command."""
    parser = argparse.ArgumentParser(
        prog="kct netlist",
        description="Netlist analysis and comparison tools",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=__doc__,
    )
    subparsers = parser.add_subparsers(dest="command", help="Netlist commands")

    # analyze subcommand
    analyze_parser = subparsers.add_parser("analyze", help="Show connectivity statistics")
    analyze_parser.add_argument("schematic", help="Path to .kicad_sch file")
    analyze_parser.add_argument(
        "--format", choices=["text", "json"], default="text", help="Output format"
    )

    # list subcommand
    list_parser = subparsers.add_parser("list", help="List all nets with connection counts")
    list_parser.add_argument("schematic", help="Path to .kicad_sch file")
    list_parser.add_argument(
        "--format", choices=["table", "json"], default="table", help="Output format"
    )
    list_parser.add_argument(
        "--sort",
        choices=["name", "connections"],
        default="connections",
        help="Sort order (default: connections)",
    )

    # show subcommand
    show_parser = subparsers.add_parser("show", help="Show specific net details")
    show_parser.add_argument("schematic", help="Path to .kicad_sch file")
    show_parser.add_argument("--net", required=True, help="Net name to show")
    show_parser.add_argument(
        "--format", choices=["text", "json"], default="text", help="Output format"
    )

    # check subcommand
    check_parser = subparsers.add_parser("check", help="Find connectivity issues")
    check_parser.add_argument("schematic", help="Path to .kicad_sch file")
    check_parser.add_argument(
        "--format", choices=["text", "json"], default="text", help="Output format"
    )

    # compare subcommand
    compare_parser = subparsers.add_parser("compare", help="Compare two netlists")
    compare_parser.add_argument("old", help="Path to old .kicad_sch file")
    compare_parser.add_argument("new", help="Path to new .kicad_sch file")
    compare_parser.add_argument(
        "--format", choices=["text", "json"], default="text", help="Output format"
    )

    # export subcommand
    export_parser = subparsers.add_parser("export", help="Export netlist file")
    export_parser.add_argument("schematic", help="Path to .kicad_sch file")
    export_parser.add_argument("-o", "--output", help="Output file path")
    export_parser.add_argument(
        "--format",
        choices=["kicad", "json"],
        default="kicad",
        help="Output format (default: kicad)",
    )

    args = parser.parse_args(argv)

    if not args.command:
        parser.print_help()
        return 0

    try:
        if args.command == "analyze":
            return cmd_analyze(Path(args.schematic), args.format)
        elif args.command == "list":
            return cmd_list(Path(args.schematic), args.format, args.sort)
        elif args.command == "show":
            return cmd_show(Path(args.schematic), args.net, args.format)
        elif args.command == "check":
            return cmd_check(Path(args.schematic), args.format)
        elif args.command == "compare":
            return cmd_compare(Path(args.old), Path(args.new), args.format)
        elif args.command == "export":
            return cmd_export(
                Path(args.schematic),
                Path(args.output) if args.output else None,
                args.format,
            )
    except FileNotFoundError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1
    except RuntimeError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1

    return 0


def cmd_analyze(schematic_path: Path, format: str) -> int:
    """Show connectivity statistics."""
    netlist = export_netlist(schematic_path)
    stats = netlist.summary()

    # Additional analysis
    single_pin_nets = find_single_pin_nets(netlist)
    stats["single_pin_net_count"] = len(single_pin_nets)

    if format == "json":
        print(json.dumps(stats, indent=2))
    else:
        print_analyze_text(stats, netlist)

    return 0


def print_analyze_text(stats: dict, netlist: Netlist) -> None:
    """Print analyze output in text format."""
    print("=" * 60)
    print("NETLIST ANALYSIS")
    print("=" * 60)

    if stats.get("source_file"):
        print(f"Source: {Path(stats['source_file']).name}")
    if stats.get("tool"):
        print(f"Tool: {stats['tool']}")
    if stats.get("date"):
        print(f"Date: {stats['date']}")

    print(f"\nSheets: {stats['sheet_count']}")
    print(f"Components: {stats['component_count']}")

    if stats.get("components_by_type"):
        print("  By type:")
        for prefix, count in sorted(stats["components_by_type"].items()):
            print(f"    {prefix}: {count}")

    print(f"\nNets: {stats['net_count']}")
    print(f"  Power: {stats['power_net_count']}")
    print(f"  Signal: {stats['signal_net_count']}")

    single_pin_count = stats.get("single_pin_net_count", 0)
    if single_pin_count > 0:
        print(f"\n⚠ Single-pin nets: {single_pin_count} (potential issues)")
    else:
        print("\n✓ No single-pin nets (connectivity looks good)")

    print("=" * 60)


def cmd_list(schematic_path: Path, format: str, sort: str) -> int:
    """List all nets with connection counts."""
    netlist = export_netlist(schematic_path)

    if sort == "connections":
        sorted_nets = sorted(netlist.nets, key=lambda n: -n.connection_count)
    else:
        sorted_nets = sorted(netlist.nets, key=lambda n: n.name)

    if format == "json":
        print_list_json(sorted_nets, netlist)
    else:
        print_list_table(sorted_nets, netlist)

    return 0


def print_list_json(nets: list[NetlistNet], netlist: Netlist) -> None:
    """Print net list as JSON."""
    power_nets = {n.name for n in netlist.power_nets}
    data = []
    for net in nets:
        nodes_summary = [f"{node.reference}.{node.pin}" for node in net.nodes[:5]]
        if len(net.nodes) > 5:
            nodes_summary.append(f"... +{len(net.nodes) - 5} more")
        data.append(
            {
                "name": net.name,
                "connections": net.connection_count,
                "type": "power" if net.name in power_nets else "signal",
                "pins": nodes_summary,
            }
        )
    print(json.dumps(data, indent=2))


def print_list_table(nets: list[NetlistNet], netlist: Netlist) -> None:
    """Print net list as table."""
    if not nets:
        print("No nets found.")
        return

    power_nets = {n.name for n in netlist.power_nets}

    print(f"{'Name':<30} {'Conn':<6} {'Type':<8} Connected Pins")
    print("-" * 80)

    for net in nets:
        net_type = "power" if net.name in power_nets else "signal"
        pins = [f"{node.reference}.{node.pin}" for node in net.nodes[:4]]
        pins_str = ", ".join(pins)
        if len(net.nodes) > 4:
            pins_str += f" +{len(net.nodes) - 4} more"

        # Truncate name if too long
        name = net.name[:29] if len(net.name) > 29 else net.name
        print(f"{name:<30} {net.connection_count:<6} {net_type:<8} {pins_str}")

    print(f"\nTotal: {len(nets)} nets")


def cmd_show(schematic_path: Path, net_name: str, format: str) -> int:
    """Show specific net details."""
    netlist = export_netlist(schematic_path)
    net = netlist.get_net(net_name)

    if not net:
        # Try fuzzy match
        similar = find_similar_nets(netlist, net_name)
        print(f"Error: Net '{net_name}' not found", file=sys.stderr)
        if similar:
            print(f"Did you mean: {', '.join(similar[:5])}", file=sys.stderr)
        return 1

    if format == "json":
        print_show_json(net, netlist)
    else:
        print_show_text(net, netlist)

    return 0


def find_similar_nets(netlist: Netlist, name: str) -> list[str]:
    """Find net names similar to the given name."""
    name_lower = name.lower()
    matches = []
    for net in netlist.nets:
        if name_lower in net.name.lower():
            matches.append(net.name)
    return sorted(matches)[:10]


def print_show_json(net: NetlistNet, netlist: Netlist) -> None:
    """Print net details as JSON."""
    power_nets = {n.name for n in netlist.power_nets}
    data = {
        "name": net.name,
        "code": net.code,
        "connections": net.connection_count,
        "type": "power" if net.name in power_nets else "signal",
        "nodes": [
            {
                "reference": node.reference,
                "pin": node.pin,
                "pin_function": node.pin_function,
                "pin_type": node.pin_type,
            }
            for node in net.nodes
        ],
    }
    print(json.dumps(data, indent=2))


def print_show_text(net: NetlistNet, netlist: Netlist) -> None:
    """Print net details as text."""
    power_nets = {n.name for n in netlist.power_nets}
    net_type = "power" if net.name in power_nets else "signal"

    print(f"Net: {net.name}")
    print("=" * 50)
    print(f"Code: {net.code}")
    print(f"Type: {net_type}")
    print(f"Connections: {net.connection_count}")

    print("\nConnected Pins:")
    print("-" * 50)
    for node in net.nodes:
        func = f" ({node.pin_function})" if node.pin_function else ""
        ptype = f" [{node.pin_type}]" if node.pin_type else ""
        print(f"  {node.reference}.{node.pin}{func}{ptype}")


def cmd_check(schematic_path: Path, format: str) -> int:
    """Find connectivity issues."""
    netlist = export_netlist(schematic_path)

    # Find issues
    single_pin_nets = find_single_pin_nets(netlist)
    power_nets = netlist.power_nets

    issues: list[dict] = []
    for net in single_pin_nets:
        node = net.nodes[0]
        issues.append(
            {
                "type": "single_pin_net",
                "severity": "warning",
                "net": net.name,
                "reference": node.reference,
                "pin": node.pin,
                "message": f"Net '{net.name}' has only 1 connection ({node.reference}.{node.pin})",
            }
        )

    result = {
        "total_nets": len(netlist.nets),
        "power_nets": len(power_nets),
        "single_pin_nets": len(single_pin_nets),
        "issues": issues,
        "power_net_status": [
            {"name": n.name, "connections": n.connection_count, "status": "ok"} for n in power_nets
        ],
    }

    if format == "json":
        print(json.dumps(result, indent=2))
    else:
        print_check_text(result, power_nets)

    return 0


def print_check_text(result: dict, power_nets: list[NetlistNet]) -> None:
    """Print check results as text."""
    print("=" * 60)
    print("NETLIST CHECK RESULTS")
    print("=" * 60)

    issues = result.get("issues", [])
    if issues:
        print(f"\n⚠ Potential Issues ({len(issues)}):")
        print("-" * 60)
        for issue in issues:
            print(f"  {issue['severity'].upper()}: {issue['message']}")
    else:
        print("\n✓ No connectivity issues found")

    print(f"\nPower Net Status ({len(power_nets)}):")
    print("-" * 60)
    for net in sorted(power_nets, key=lambda n: n.name):
        print(f"  ✓ {net.name} ({net.connection_count} connections)")

    print("\nSummary:")
    print(f"  Total nets: {result['total_nets']}")
    print(f"  Power nets: {result['power_nets']}")
    print(f"  Single-pin nets: {result['single_pin_nets']}")
    print("=" * 60)


def cmd_compare(old_path: Path, new_path: Path, format: str) -> int:
    """Compare two netlists."""
    old_netlist = export_netlist(old_path)
    new_netlist = export_netlist(new_path)

    # Compare components
    old_refs = {c.reference for c in old_netlist.components}
    new_refs = {c.reference for c in new_netlist.components}
    added_components = sorted(new_refs - old_refs)
    removed_components = sorted(old_refs - new_refs)

    # Compare nets
    old_nets = {n.name: n for n in old_netlist.nets}
    new_nets = {n.name: n for n in new_netlist.nets}
    added_nets = sorted(set(new_nets.keys()) - set(old_nets.keys()))
    removed_nets = sorted(set(old_nets.keys()) - set(new_nets.keys()))

    # Find modified nets (same name but different connections)
    modified_nets = []
    for name in set(old_nets.keys()) & set(new_nets.keys()):
        old_count = old_nets[name].connection_count
        new_count = new_nets[name].connection_count
        if old_count != new_count:
            diff = new_count - old_count
            diff_str = f"+{diff}" if diff > 0 else str(diff)
            modified_nets.append(
                {"name": name, "old": old_count, "new": new_count, "diff": diff_str}
            )

    result = {
        "old_file": str(old_path),
        "new_file": str(new_path),
        "components": {
            "added": added_components,
            "removed": removed_components,
            "added_count": len(added_components),
            "removed_count": len(removed_components),
        },
        "nets": {
            "added": added_nets,
            "removed": removed_nets,
            "modified": modified_nets,
            "added_count": len(added_nets),
            "removed_count": len(removed_nets),
            "modified_count": len(modified_nets),
        },
    }

    if format == "json":
        print(json.dumps(result, indent=2))
    else:
        print_compare_text(result)

    return 0


def print_compare_text(result: dict) -> None:
    """Print comparison results as text."""
    print("=" * 60)
    print("NETLIST COMPARISON")
    print("=" * 60)
    print(f"Old: {Path(result['old_file']).name}")
    print(f"New: {Path(result['new_file']).name}")

    comps = result["components"]
    print("\nComponents:")
    print("-" * 40)
    if comps["added"]:
        print(f"  Added ({comps['added_count']}): {', '.join(comps['added'][:10])}")
        if comps["added_count"] > 10:
            print(f"         ... +{comps['added_count'] - 10} more")
    if comps["removed"]:
        print(f"  Removed ({comps['removed_count']}): {', '.join(comps['removed'][:10])}")
        if comps["removed_count"] > 10:
            print(f"           ... +{comps['removed_count'] - 10} more")
    if not comps["added"] and not comps["removed"]:
        print("  No component changes")

    nets = result["nets"]
    print("\nNets:")
    print("-" * 40)
    if nets["added"]:
        print(f"  Added ({nets['added_count']}): {', '.join(nets['added'][:10])}")
        if nets["added_count"] > 10:
            print(f"         ... +{nets['added_count'] - 10} more")
    if nets["removed"]:
        print(f"  Removed ({nets['removed_count']}): {', '.join(nets['removed'][:10])}")
        if nets["removed_count"] > 10:
            print(f"           ... +{nets['removed_count'] - 10} more")
    if nets["modified"]:
        print(f"  Modified ({nets['modified_count']}):")
        for mod in nets["modified"][:10]:
            print(f"    {mod['name']}: {mod['old']} → {mod['new']} ({mod['diff']})")
        if nets["modified_count"] > 10:
            print(f"    ... +{nets['modified_count'] - 10} more")
    if not nets["added"] and not nets["removed"] and not nets["modified"]:
        print("  No net changes")

    print("=" * 60)


def cmd_export(schematic_path: Path, output_path: Path | None, format: str) -> int:
    """Export netlist file."""
    netlist = export_netlist(schematic_path)

    if format == "json":
        output = netlist.to_json()
        if output_path:
            output_path.write_text(output)
            print(f"Exported JSON netlist to: {output_path}")
        else:
            print(output)
    else:
        # KiCad format - the netlist is already exported by export_netlist()
        # We just need to tell the user where it is
        default_output = schematic_path.parent / f"{schematic_path.stem}-netlist.kicad_net"
        if output_path and output_path != default_output:
            # Copy to requested location
            import shutil

            shutil.copy(default_output, output_path)
            print(f"Exported KiCad netlist to: {output_path}")
        else:
            print(f"Exported KiCad netlist to: {default_output}")

    return 0


def find_single_pin_nets(netlist: Netlist) -> list[NetlistNet]:
    """Find nets with only one connection (potential issues)."""
    return [net for net in netlist.nets if net.connection_count == 1]


if __name__ == "__main__":
    sys.exit(main())



"""
Evolutionary (genetic algorithm) placement optimizer.

Provides global optimization for component placement using:
- Population-based search with tournament selection
- Spatial partitioning crossover
- Gaussian mutation for positions and rotation perturbation
- Multi-objective fitness function (wire length, conflicts, routability)

Example::

    from kicad_tools.optim.evolutionary import EvolutionaryPlacementOptimizer
    from kicad_tools.schema.pcb import PCB

    pcb = PCB.load("board.kicad_pcb")
    optimizer = EvolutionaryPlacementOptimizer.from_pcb(pcb)

    # Run evolutionary optimization
    best = optimizer.optimize(generations=100, population_size=50)

    # Or use hybrid mode: evolutionary + physics refinement
    physics_optimizer = optimizer.optimize_hybrid(generations=50)
    physics_optimizer.write_to_pcb(pcb)
    pcb.save("optimized.kicad_pcb")
"""

from __future__ import annotations

import math
import random
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Callable

from kicad_tools.optim.components import Component, FunctionalCluster, Spring
from kicad_tools.optim.config import PlacementConfig
from kicad_tools.optim.geometry import Polygon, Vector2D
from kicad_tools.optim.placement import PlacementOptimizer

if TYPE_CHECKING:
    from kicad_tools.schema.pcb import PCB

__all__ = [
    "EvolutionaryPlacementOptimizer",
    "EvolutionaryConfig",
    "Individual",
]


@dataclass
class EvolutionaryConfig:
    """Configuration for the evolutionary placement optimizer."""

    # Population parameters
    population_size: int = 50
    generations: int = 100
    elitism: int = 5  # Keep top N individuals unchanged each generation

    # Genetic operator rates
    crossover_rate: float = 0.8
    mutation_rate: float = 0.1
    tournament_size: int = 3

    # Mutation parameters
    position_mutation_sigma: float = 1.0  # Standard deviation in mm for position mutations
    rotation_mutation_prob: float = 0.05  # Probability of rotating a component 90 degrees

    # Fitness weights (higher = more important)
    wire_length_weight: float = 0.1
    conflict_weight: float = 100.0
    routability_weight: float = 50.0
    boundary_violation_weight: float = 500.0  # Heavy penalty for components outside board

    # Grid snapping
    grid_snap: float = 0.127  # 5 mil grid (0 to disable)
    rotation_snap: float = 90.0  # Rotation grid in degrees

    # Convergence detection
    convergence_generations: int = 20  # Check convergence over this many generations
    convergence_threshold: float = 0.001  # Relative improvement threshold

    # Parallel processing
    parallel: bool = True
    max_workers: int | None = None  # None = use all available cores


@dataclass
class Individual:
    """
    A placement configuration (chromosome) in the genetic algorithm.

    Stores component positions and rotations as the genotype.
    Fitness is computed based on wire length, conflicts, and routability.
    """

    positions: dict[str, tuple[float, float]] = field(default_factory=dict)  # ref -> (x, y)
    rotations: dict[str, float] = field(default_factory=dict)  # ref -> rotation (degrees)
    fitness: float = 0.0

    def copy(self) -> Individual:
        """Create a deep copy of this individual."""
        return Individual(
            positions=dict(self.positions),
            rotations=dict(self.rotations),
            fitness=self.fitness,
        )


class EvolutionaryPlacementOptimizer:
    """
    Genetic algorithm for component placement optimization.

    Uses population-based search to explore the placement space globally,
    avoiding local minima that physics-based simulation might get stuck in.

    Key features:
    - Tournament selection for parent selection
    - Spatial partitioning crossover to preserve locality
    - Gaussian mutation for position exploration
    - Multi-objective fitness function
    - Optional parallel fitness evaluation
    """

    def __init__(
        self,
        board_outline: Polygon,
        config: EvolutionaryConfig | None = None,
    ):
        """
        Initialize the evolutionary optimizer.

        Args:
            board_outline: Polygon defining the board boundary
            config: Optimization parameters
        """
        self.board_outline = board_outline
        self.config = config or EvolutionaryConfig()
        self.components: list[Component] = []
        self.springs: list[Spring] = []
        self.clusters: list[FunctionalCluster] = []
        self._component_map: dict[str, Component] = {}
        self._board_bounds = self._compute_board_bounds()
        self._fitness_history: list[float] = []
        self._cluster_members: set[str] = set()  # Components that are in clusters

    def _compute_board_bounds(self) -> tuple[float, float, float, float]:
        """Compute bounding box of board outline (min_x, min_y, max_x, max_y)."""
        if not self.board_outline.vertices:
            return (0.0, 0.0, 100.0, 100.0)
        xs = [v.x for v in self.board_outline.vertices]
        ys = [v.y for v in self.board_outline.vertices]
        return (min(xs), min(ys), max(xs), max(ys))

    @classmethod
    def from_pcb(
        cls,
        pcb: PCB,
        config: EvolutionaryConfig | None = None,
        fixed_refs: list[str] | None = None,
        enable_clustering: bool = False,
    ) -> EvolutionaryPlacementOptimizer:
        """
        Create optimizer from a loaded PCB.

        Delegates to PlacementOptimizer.from_pcb() to reuse component
        and spring extraction logic, then copies the data.

        Args:
            pcb: Loaded PCB object
            config: Optimization parameters
            fixed_refs: List of reference designators for fixed components
            enable_clustering: If True, detect and preserve functional clusters
        """
        # Use PlacementOptimizer to extract components and nets
        physics_opt = PlacementOptimizer.from_pcb(
            pcb, fixed_refs=fixed_refs, enable_clustering=enable_clustering
        )

        optimizer = cls(physics_opt.board_outline, config)
        optimizer.components = physics_opt.components
        optimizer.springs = physics_opt.springs
        optimizer.clusters = physics_opt.clusters
        optimizer._component_map = physics_opt._component_map
        optimizer._update_cluster_members()

        return optimizer

    @classmethod
    def from_placement_optimizer(
        cls,
        physics_optimizer: PlacementOptimizer,
        config: EvolutionaryConfig | None = None,
    ) -> EvolutionaryPlacementOptimizer:
        """
        Create evolutionary optimizer from an existing PlacementOptimizer.

        Useful for hybrid optimization workflows.
        """
        optimizer = cls(physics_optimizer.board_outline, config)
        optimizer.components = physics_optimizer.components
        optimizer.springs = physics_optimizer.springs
        optimizer.clusters = physics_optimizer.clusters
        optimizer._component_map = physics_optimizer._component_map
        optimizer._update_cluster_members()

        return optimizer

    def _update_cluster_members(self):
        """Build set of all component refs that are part of clusters."""
        self._cluster_members = set()
        for cluster in self.clusters:
            self._cluster_members.add(cluster.anchor)
            self._cluster_members.update(cluster.members)

    def add_component(self, comp: Component):
        """Add a component to the optimizer."""
        self.components.append(comp)
        self._component_map[comp.ref] = comp

    def add_spring(self, spring: Spring):
        """Add a spring (net connection) to the optimizer."""
        self.springs.append(spring)

    def _get_movable_components(self) -> list[Component]:
        """Get list of components that can be moved (not fixed)."""
        return [c for c in self.components if not c.fixed]

    def _individual_from_current(self) -> Individual:
        """Create an individual from current component positions."""
        ind = Individual()
        for comp in self._get_movable_components():
            ind.positions[comp.ref] = (comp.x, comp.y)
            ind.rotations[comp.ref] = comp.rotation
        return ind

    def _apply_individual(self, ind: Individual):
        """Apply individual's positions/rotations to components."""
        for ref, (x, y) in ind.positions.items():
            comp = self._component_map.get(ref)
            if comp and not comp.fixed:
                comp.x = x
                comp.y = y
                comp.rotation = ind.rotations.get(ref, comp.rotation)
                comp.update_pin_positions()

    def _initialize_population(self, population_size: int) -> list[Individual]:
        """
        Create initial population of random placements.

        Generates individuals with random positions within board bounds
        and random 90-degree rotations.
        """
        population = []
        movable = self._get_movable_components()
        min_x, min_y, max_x, max_y = self._board_bounds

        # Add margin to keep components inside
        margin = 2.0
        min_x += margin
        min_y += margin
        max_x -= margin
        max_y -= margin

        # First individual preserves current placement
        population.append(self._individual_from_current())

        # Generate random individuals
        for _ in range(population_size - 1):
            ind = Individual()
            for comp in movable:
                # Random position within board bounds
                x = random.uniform(min_x, max_x)
                y = random.uniform(min_y, max_y)

                # Snap to grid if configured
                if self.config.grid_snap > 0:
                    x = round(x / self.config.grid_snap) * self.config.grid_snap
                    y = round(y / self.config.grid_snap) * self.config.grid_snap

                ind.positions[comp.ref] = (x, y)

                # Random 90-degree rotation
                rotation = random.choice([0.0, 90.0, 180.0, 270.0])
                ind.rotations[comp.ref] = rotation

            population.append(ind)

        return population

    def _crossover(self, parent1: Individual, parent2: Individual) -> Individual:
        """
        Crossover: exchange component subsets between parents.

        Uses spatial partitioning - components in left half from parent1,
        right half from parent2 (with some randomization).

        Cluster integrity is preserved: all members of a cluster are taken
        from the same parent as the cluster anchor.
        """
        child = Individual()
        min_x, _, max_x, _ = self._board_bounds
        mid_x = (min_x + max_x) / 2

        # Add some randomization to the partition line
        partition_x = mid_x + random.gauss(0, (max_x - min_x) * 0.1)

        # Determine which parent each cluster should come from (based on anchor position)
        cluster_parent: dict[str, Individual] = {}  # ref -> parent for cluster members
        for cluster in self.clusters:
            if cluster.anchor in parent1.positions:
                anchor_x, _ = parent1.positions[cluster.anchor]
                chosen_parent = parent1 if anchor_x < partition_x else parent2
                # All cluster members should come from same parent
                cluster_parent[cluster.anchor] = chosen_parent
                for member in cluster.members:
                    cluster_parent[member] = chosen_parent

        for ref in parent1.positions:
            # Check if this component is part of a cluster
            if ref in cluster_parent:
                chosen = cluster_parent[ref]
                child.positions[ref] = chosen.positions[ref]
                child.rotations[ref] = chosen.rotations[ref]
            else:
                # Use parent1's current position to decide partition
                p1_x, p1_y = parent1.positions[ref]

                if p1_x < partition_x:
                    # Take from parent1
                    child.positions[ref] = parent1.positions[ref]
                    child.rotations[ref] = parent1.rotations[ref]
                else:
                    # Take from parent2
                    child.positions[ref] = parent2.positions[ref]
                    child.rotations[ref] = parent2.rotations[ref]

        return child

    def _mutate(self, ind: Individual) -> Individual:
        """
        Mutation: random position/rotation perturbations.

        - Position: Gaussian noise with configured sigma
        - Rotation: 90-degree increment with low probability
        """
        min_x, min_y, max_x, max_y = self._board_bounds
        margin = 1.0

        for ref in ind.positions:
            # Position mutation
            if random.random() < self.config.mutation_rate:
                x, y = ind.positions[ref]
                x += random.gauss(0, self.config.position_mutation_sigma)
                y += random.gauss(0, self.config.position_mutation_sigma)

                # Clamp to board bounds
                x = max(min_x + margin, min(max_x - margin, x))
                y = max(min_y + margin, min(max_y - margin, y))

                # Snap to grid if configured
                if self.config.grid_snap > 0:
                    x = round(x / self.config.grid_snap) * self.config.grid_snap
                    y = round(y / self.config.grid_snap) * self.config.grid_snap

                ind.positions[ref] = (x, y)

            # Rotation mutation
            if random.random() < self.config.rotation_mutation_prob:
                current_rot = ind.rotations[ref]
                new_rot = (current_rot + 90) % 360
                ind.rotations[ref] = new_rot

        return ind

    def _evaluate_fitness(self, ind: Individual) -> float:
        """
        Multi-objective fitness combining wire length, conflicts, and routability.

        Higher fitness is better.
        """
        # Apply positions temporarily
        original_positions = {}
        for ref, (x, y) in ind.positions.items():
            comp = self._component_map.get(ref)
            if comp:
                original_positions[ref] = (comp.x, comp.y, comp.rotation)
                comp.x = x
                comp.y = y
                comp.rotation = ind.rotations.get(ref, comp.rotation)
                comp.update_pin_positions()

        try:
            # Calculate objectives
            wire_length = self._total_wire_length()
            conflicts = self._count_conflicts()
            boundary_violations = self._count_boundary_violations()
            routability = self._estimate_routability()

            # Weighted sum (higher = better)
            fitness = (
                1000.0
                - wire_length * self.config.wire_length_weight
                - conflicts * self.config.conflict_weight
                - boundary_violations * self.config.boundary_violation_weight
                + routability * self.config.routability_weight
            )

            return fitness

        finally:
            # Restore original positions
            for ref, (x, y, rot) in original_positions.items():
                comp = self._component_map.get(ref)
                if comp:
                    comp.x = x
                    comp.y = y
                    comp.rotation = rot
                    comp.update_pin_positions()

    def _total_wire_length(self) -> float:
        """Compute total wire length from spring connections."""
        total = 0.0
        for spring in self.springs:
            comp1 = self._component_map.get(spring.comp1_ref)
            comp2 = self._component_map.get(spring.comp2_ref)

            if not comp1 or not comp2:
                continue

            pin1 = next((p for p in comp1.pins if p.number == spring.pin1_num), None)
            pin2 = next((p for p in comp2.pins if p.number == spring.pin2_num), None)

            if not pin1 or not pin2:
                continue

            dx = pin2.x - pin1.x
            dy = pin2.y - pin1.y
            total += math.sqrt(dx * dx + dy * dy)

        return total

    def _count_conflicts(self) -> int:
        """
        Count component overlaps (courtyard conflicts).

        Uses axis-aligned bounding box overlap detection.
        """
        conflicts = 0
        n = len(self.components)

        for i in range(n):
            comp1 = self.components[i]
            # Get AABB for comp1
            hw1, hh1 = comp1.width / 2, comp1.height / 2

            for j in range(i + 1, n):
                comp2 = self.components[j]
                hw2, hh2 = comp2.width / 2, comp2.height / 2

                # AABB overlap check
                dx = abs(comp1.x - comp2.x)
                dy = abs(comp1.y - comp2.y)

                if dx < (hw1 + hw2) and dy < (hh1 + hh2):
                    conflicts += 1

        return conflicts

    def _count_boundary_violations(self) -> int:
        """Count components that are outside board boundary."""
        violations = 0
        for comp in self.components:
            if not self.board_outline.contains_point(Vector2D(comp.x, comp.y)):
                violations += 1
        return violations

    def _estimate_routability(self) -> float:
        """
        Estimate routability based on component spacing and wire congestion.

        Returns a score from 0-100 where higher is better.
        """
        if not self.components:
            return 100.0

        # Calculate average spacing between components
        n = len(self.components)
        if n < 2:
            return 100.0

        total_spacing = 0.0
        count = 0
        for i in range(n):
            for j in range(i + 1, n):
                dx = self.components[i].x - self.components[j].x
                dy = self.components[i].y - self.components[j].y
                dist = math.sqrt(dx * dx + dy * dy)
                total_spacing += dist
                count += 1

        avg_spacing = total_spacing / count if count > 0 else 0

        # More spacing = better routability (up to a point)
        # Normalize to ~0-100 range
        routability_score = min(100.0, avg_spacing * 5.0)

        return routability_score

    def _tournament_select(self, population: list[Individual]) -> Individual:
        """Select individual using tournament selection."""
        tournament = random.sample(population, min(self.config.tournament_size, len(population)))
        return max(tournament, key=lambda ind: ind.fitness)

    def _evolve(self, population: list[Individual]) -> list[Individual]:
        """
        Perform one generation of evolution.

        Uses elitism, tournament selection, crossover, and mutation.
        """
        # Sort by fitness (highest first)
        population.sort(key=lambda ind: ind.fitness, reverse=True)

        new_population = []

        # Elitism: keep top individuals unchanged
        for i in range(min(self.config.elitism, len(population))):
            new_population.append(population[i].copy())

        # Fill rest with offspring
        while len(new_population) < len(population):
            # Tournament selection for parents
            parent1 = self._tournament_select(population)
            parent2 = self._tournament_select(population)

            # Crossover
            if random.random() < self.config.crossover_rate:
                child = self._crossover(parent1, parent2)
            else:
                child = parent1.copy()

            # Mutation
            child = self._mutate(child)

            new_population.append(child)

        return new_population

    def _evaluate_population(self, population: list[Individual]):
        """
        Evaluate fitness for all individuals in population.

        Uses parallel processing if configured.
        """
        if self.config.parallel and len(population) > 4:
            # Parallel evaluation
            # Note: This requires picklable objects; for simplicity,
            # we'll evaluate sequentially in the current implementation
            # since Component objects have complex state.
            # A proper implementation would serialize the individual data.
            for ind in population:
                ind.fitness = self._evaluate_fitness(ind)
        else:
            # Sequential evaluation
            for ind in population:
                ind.fitness = self._evaluate_fitness(ind)

    def _check_convergence(self) -> bool:
        """
        Check if population has converged.

        Returns True if fitness improvement has plateaued.
        """
        if len(self._fitness_history) < self.config.convergence_generations:
            return False

        recent = self._fitness_history[-self.config.convergence_generations :]
        if recent[0] == 0:
            return False

        improvement = (recent[-1] - recent[0]) / abs(recent[0])
        return improvement < self.config.convergence_threshold

    def optimize(
        self,
        generations: int | None = None,
        population_size: int | None = None,
        callback: Callable[[int, Individual], None] | None = None,
    ) -> Individual:
        """
        Run evolutionary optimization.

        Args:
            generations: Number of generations (default: config.generations)
            population_size: Population size (default: config.population_size)
            callback: Optional function called each generation with (gen, best_individual)

        Returns:
            Best individual found
        """
        generations = generations or self.config.generations
        population_size = population_size or self.config.population_size

        # Initialize population
        population = self._initialize_population(population_size)
        self._fitness_history = []

        for gen in range(generations):
            # Evaluate fitness
            self._evaluate_population(population)

            # Sort by fitness (highest first)
            population.sort(key=lambda ind: ind.fitness, reverse=True)
            best = population[0]

            self._fitness_history.append(best.fitness)

            if callback:
                callback(gen, best)

            # Check convergence
            if self._check_convergence():
                break

            # Evolve to next generation (except on last iteration)
            if gen < generations - 1:
                population = self._evolve(population)

        # Return best individual
        return population[0]

    def optimize_hybrid(
        self,
        evolutionary_generations: int = 50,
        population_size: int = 30,
        physics_iterations: int = 500,
        physics_config: PlacementConfig | None = None,
        callback: Callable[[int, Individual], None] | None = None,
    ) -> PlacementOptimizer:
        """
        Hybrid optimization: evolutionary global search + physics local refinement.

        Phase 1: Run evolutionary algorithm to find globally good placement
        Phase 2: Use physics-based simulation to refine the best result

        Args:
            evolutionary_generations: Generations for evolutionary phase
            population_size: Population size for evolutionary phase
            physics_iterations: Iterations for physics refinement
            physics_config: Configuration for physics optimizer
            callback: Optional callback for evolutionary phase

        Returns:
            PlacementOptimizer with refined placement (call write_to_pcb to save)
        """
        # Phase 1: Evolutionary optimization
        best = self.optimize(
            generations=evolutionary_generations,
            population_size=population_size,
            callback=callback,
        )

        # Apply best individual to components
        self._apply_individual(best)

        # Phase 2: Create physics optimizer from current state
        physics_opt = PlacementOptimizer(self.board_outline, physics_config)
        physics_opt.components = self.components
        physics_opt.springs = self.springs
        physics_opt._component_map = self._component_map

        # Run physics simulation for local refinement
        physics_opt.run(iterations=physics_iterations)

        # Snap to grid
        if self.config.grid_snap > 0:
            physics_opt.snap_positions(self.config.grid_snap)
        if self.config.rotation_snap > 0:
            physics_opt.snap_rotations(self.config.rotation_snap)

        return physics_opt

    def write_to_pcb(self, pcb: PCB) -> int:
        """
        Write current component positions back to a PCB object.

        Args:
            pcb: PCB object to update

        Returns:
            Number of components successfully updated
        """
        updated = 0
        for comp in self.components:
            if pcb.update_footprint_position(comp.ref, comp.x, comp.y, comp.rotation):
                updated += 1
        return updated

    def report(self) -> str:
        """Generate a text report of current placement."""
        wire_length = self._total_wire_length()
        conflicts = self._count_conflicts()
        boundary_violations = self._count_boundary_violations()

        lines = [
            "Evolutionary Placement Optimizer Report",
            "=" * 45,
            f"Components: {len(self.components)}",
            f"Springs (net connections): {len(self.springs)}",
            f"Total wire length: {wire_length:.2f} mm",
            f"Conflicts (overlaps): {conflicts}",
            f"Boundary violations: {boundary_violations}",
            "",
            "Component Positions:",
            "-" * 45,
        ]

        for comp in sorted(self.components, key=lambda c: c.ref):
            fixed = " [fixed]" if comp.fixed else ""
            lines.append(
                f"  {comp.ref:8s}: ({comp.x:7.2f}, {comp.y:7.2f}) @ {comp.rotation:6.1f}°{fixed}"
            )

        return "\n".join(lines)



"""ERC (Electrical Rules Check) report parsing and validation.

This module provides tools for parsing KiCad ERC reports and analyzing
schematic electrical rule violations.

Example:
    >>> from kicad_tools.erc import ERCReport, ERCViolation
    >>> report = ERCReport.load("design-erc.json")
    >>> print(f"Found {report.error_count} errors")
    >>> for v in report.errors:
    ...     print(f"  {v.type}: {v.description}")
"""

from .report import (
    ERCReport,
    parse_json_report,
    parse_text_report,
)
from .violation import (
    ERC_CATEGORIES,
    ERC_TYPE_DESCRIPTIONS,
    ERCViolation,
    ERCViolationType,
    Severity,
)

__all__ = [
    # Violation types
    "ERCViolation",
    "ERCViolationType",
    "Severity",
    "ERC_TYPE_DESCRIPTIONS",
    "ERC_CATEGORIES",
    # Report parsing
    "ERCReport",
    "parse_json_report",
    "parse_text_report",
]



"""Tests for PCB-related CLI commands."""

import json
from pathlib import Path

import pytest


class TestPcbQuery:
    """Tests for pcb_query.py CLI."""

    def test_file_not_found(self, capsys, monkeypatch, tmp_path):
        """Test handling of missing file."""
        from kicad_tools.cli.pcb_query import main

        missing_file = tmp_path / "definitely_missing" / "nonexistent.kicad_pcb"
        monkeypatch.setattr("sys.argv", ["pcb-query", str(missing_file), "summary"])
        with pytest.raises(SystemExit) as exc_info:
            main()

        assert exc_info.value.code == 1
        captured = capsys.readouterr()
        assert "Error" in captured.err or "not found" in captured.err.lower()

    def test_summary_command(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test summary command."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "summary"])
        main()

        captured = capsys.readouterr()
        assert "PCB:" in captured.out
        assert len(captured.out) > 0

    def test_summary_json(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test summary command with JSON output."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr(
            "sys.argv", ["pcb-query", str(minimal_pcb), "summary", "--format", "json"]
        )
        main()

        captured = capsys.readouterr()
        data = json.loads(captured.out)
        assert isinstance(data, dict)

    def test_footprints_command(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test footprints command."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "footprints"])
        main()

        captured = capsys.readouterr()
        # Should output footprint info or "no footprints"
        assert len(captured.out) > 0

    def test_footprints_json(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test footprints command with JSON output."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr(
            "sys.argv", ["pcb-query", str(minimal_pcb), "footprints", "--format", "json"]
        )
        main()

        captured = capsys.readouterr()
        data = json.loads(captured.out)
        assert isinstance(data, list)

    def test_footprints_filter(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test footprints command with filter."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr(
            "sys.argv", ["pcb-query", str(minimal_pcb), "footprints", "--filter", "R*"]
        )
        main()

        captured = capsys.readouterr()
        # Should work without error
        assert len(captured.out) > 0

    def test_footprints_sorted(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test footprints command with sorting."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "footprints", "--sorted"])
        main()

        captured = capsys.readouterr()
        assert len(captured.out) > 0

    def test_footprint_detail(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test footprint detail command."""
        from kicad_tools.cli.pcb_query import main

        # R1 exists in minimal_pcb fixture
        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "footprint", "R1"])
        main()

        captured = capsys.readouterr()
        assert "R1" in captured.out

    def test_nets_command(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test nets command."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "nets"])
        main()

        captured = capsys.readouterr()
        assert len(captured.out) > 0

    def test_nets_json(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test nets command with JSON output."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "nets", "--format", "json"])
        main()

        captured = capsys.readouterr()
        data = json.loads(captured.out)
        assert isinstance(data, list)

    def test_nets_sorted(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test nets command with sorting."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "nets", "--sorted"])
        main()

        captured = capsys.readouterr()
        assert len(captured.out) > 0

    def test_net_detail(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test net detail command."""
        from kicad_tools.cli.pcb_query import main

        # GND exists in minimal_pcb fixture
        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "net", "GND"])
        main()

        captured = capsys.readouterr()
        assert "GND" in captured.out

    def test_traces_command(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test traces command."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "traces"])
        main()

        captured = capsys.readouterr()
        assert len(captured.out) > 0

    def test_traces_json(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test traces command with JSON output."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr(
            "sys.argv", ["pcb-query", str(minimal_pcb), "traces", "--format", "json"]
        )
        main()

        captured = capsys.readouterr()
        data = json.loads(captured.out)
        assert isinstance(data, dict)

    def test_vias_command(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test vias command."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "vias"])
        main()

        captured = capsys.readouterr()
        # May have "No vias" or via list
        assert len(captured.out) > 0

    def test_vias_json(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test vias command with JSON output."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "vias", "--format", "json"])
        main()

        captured = capsys.readouterr()
        data = json.loads(captured.out)
        assert isinstance(data, list)

    def test_stackup_command(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test stackup command."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(minimal_pcb), "stackup"])
        main()

        captured = capsys.readouterr()
        # Should show layer stackup info
        assert len(captured.out) > 0

    def test_stackup_json(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test stackup command with JSON output."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr(
            "sys.argv", ["pcb-query", str(minimal_pcb), "stackup", "--format", "json"]
        )
        main()

        captured = capsys.readouterr()
        data = json.loads(captured.out)
        assert isinstance(data, (list, dict))

    def test_routing_test_pcb(self, routing_test_pcb: Path, capsys, monkeypatch):
        """Test with routing test PCB fixture."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(routing_test_pcb), "summary"])
        main()

        captured = capsys.readouterr()
        assert "PCB:" in captured.out

    def test_zone_test_pcb(self, zone_test_pcb: Path, capsys, monkeypatch):
        """Test with zone test PCB fixture."""
        from kicad_tools.cli.pcb_query import main

        monkeypatch.setattr("sys.argv", ["pcb-query", str(zone_test_pcb), "summary"])
        main()

        captured = capsys.readouterr()
        assert "PCB:" in captured.out


class TestPcbModify:
    """Tests for pcb_modify.py CLI.

    Note: Some tests require PCBs with 'fp_text reference' format (older KiCad style).
    The minimal_pcb fixture uses 'property "Reference"' format (KiCad 8 style),
    which is not supported by the current pcb_modify implementation.
    """

    @pytest.fixture
    def legacy_pcb(self, tmp_path):
        """Create a PCB with legacy fp_text format for modify tests."""
        pcb_content = """(kicad_pcb
  (version 20231120)
  (generator "test")
  (generator_version "7.0")
  (general (thickness 1.6))
  (paper "A4")
  (layers
    (0 "F.Cu" signal)
    (31 "B.Cu" signal)
  )
  (setup (pad_to_mask_clearance 0))
  (net 0 "")
  (net 1 "GND")
  (footprint "Resistor_SMD:R_0402_1005Metric"
    (layer "F.Cu")
    (at 100 100)
    (fp_text reference "R1" (at 0 -1.5) (layer "F.SilkS"))
    (fp_text value "10k" (at 0 1.5) (layer "F.Fab"))
    (pad "1" smd roundrect (at -0.51 0) (size 0.54 0.64) (layers "F.Cu" "F.Paste" "F.Mask") (net 1 "GND"))
    (pad "2" smd roundrect (at 0.51 0) (size 0.54 0.64) (layers "F.Cu" "F.Paste" "F.Mask") (net 0 ""))
  )
  (segment (start 100 100) (end 110 100) (width 0.2) (layer "F.Cu") (net 1))
)"""
        pcb_file = tmp_path / "legacy.kicad_pcb"
        pcb_file.write_text(pcb_content)
        return pcb_file

    def test_file_not_found(self, capsys, monkeypatch, tmp_path):
        """Test handling of missing file."""
        from kicad_tools.cli.pcb_modify import main

        missing_file = tmp_path / "definitely_missing" / "nonexistent.kicad_pcb"
        monkeypatch.setattr(
            "sys.argv", ["pcb-modify", str(missing_file), "move", "R1", "100", "100"]
        )
        with pytest.raises(SystemExit) as exc_info:
            main()

        assert exc_info.value.code == 1
        captured = capsys.readouterr()
        assert "Error" in captured.err or "not found" in captured.err.lower()

    def test_move_dry_run(self, legacy_pcb: Path, capsys, monkeypatch):
        """Test move command with dry run."""
        from kicad_tools.cli.pcb_modify import main

        monkeypatch.setattr(
            "sys.argv",
            ["pcb-modify", str(legacy_pcb), "move", "R1", "120", "130", "--dry-run"],
        )
        main()

        captured = capsys.readouterr()
        assert "Moving R1" in captured.out
        assert "120" in captured.out
        assert "130" in captured.out

    def test_move_to_output_file(self, legacy_pcb: Path, tmp_path, capsys, monkeypatch):
        """Test move command writing to output file."""
        from kicad_tools.cli.pcb_modify import main

        output_file = tmp_path / "output.kicad_pcb"
        monkeypatch.setattr(
            "sys.argv",
            [
                "pcb-modify",
                str(legacy_pcb),
                "move",
                "R1",
                "120",
                "130",
                "-o",
                str(output_file),
            ],
        )
        main()

        captured = capsys.readouterr()
        assert "Moving R1" in captured.out
        assert output_file.exists()

    def test_move_footprint_not_found(self, legacy_pcb: Path, capsys, monkeypatch):
        """Test move command with non-existent footprint."""
        from kicad_tools.cli.pcb_modify import main

        monkeypatch.setattr(
            "sys.argv",
            ["pcb-modify", str(legacy_pcb), "move", "NONEXISTENT", "100", "100", "--dry-run"],
        )
        with pytest.raises(SystemExit) as exc_info:
            main()

        assert exc_info.value.code == 1
        captured = capsys.readouterr()
        assert "not found" in captured.err

    def test_rotate_dry_run(self, legacy_pcb: Path, capsys, monkeypatch):
        """Test rotate command with dry run."""
        from kicad_tools.cli.pcb_modify import main

        monkeypatch.setattr(
            "sys.argv",
            ["pcb-modify", str(legacy_pcb), "rotate", "R1", "90", "--dry-run"],
        )
        main()

        captured = capsys.readouterr()
        assert "Rotating R1" in captured.out
        assert "90" in captured.out

    def test_flip_dry_run(self, legacy_pcb: Path, capsys, monkeypatch):
        """Test flip command with dry run."""
        from kicad_tools.cli.pcb_modify import main

        monkeypatch.setattr("sys.argv", ["pcb-modify", str(legacy_pcb), "flip", "R1", "--dry-run"])
        main()

        captured = capsys.readouterr()
        assert "Flipping R1" in captured.out

    def test_update_value_dry_run(self, legacy_pcb: Path, capsys, monkeypatch):
        """Test update-value command with dry run."""
        from kicad_tools.cli.pcb_modify import main

        monkeypatch.setattr(
            "sys.argv",
            ["pcb-modify", str(legacy_pcb), "update-value", "R1", "22k", "--dry-run"],
        )
        main()

        captured = capsys.readouterr()
        assert "Updating value" in captured.out or "R1" in captured.out

    def test_rename_dry_run(self, legacy_pcb: Path, capsys, monkeypatch):
        """Test rename command with dry run."""
        from kicad_tools.cli.pcb_modify import main

        monkeypatch.setattr(
            "sys.argv",
            ["pcb-modify", str(legacy_pcb), "rename", "R1", "R10", "--dry-run"],
        )
        main()

        captured = capsys.readouterr()
        assert "Renaming" in captured.out or "R1" in captured.out

    def test_delete_traces_dry_run(self, legacy_pcb: Path, capsys, monkeypatch):
        """Test delete-traces command with dry run."""
        from kicad_tools.cli.pcb_modify import main

        monkeypatch.setattr(
            "sys.argv",
            ["pcb-modify", str(legacy_pcb), "delete-traces", "GND", "--dry-run"],
        )
        main()

        captured = capsys.readouterr()
        # Should show "Deleting" or "deleted" info
        assert "Deleting" in captured.out or "deleted" in captured.out.lower()

    def test_kicad8_pcb_footprint_not_found(self, minimal_pcb: Path, capsys, monkeypatch):
        """Test that KiCad 8 PCBs with 'property' format report footprint not found.

        This is a known limitation - the pcb_modify CLI expects 'fp_text reference'
        but KiCad 8 uses 'property "Reference"' format.
        """
        from kicad_tools.cli.pcb_modify import main

        monkeypatch.setattr(
            "sys.argv",
            ["pcb-modify", str(minimal_pcb), "move", "R1", "100", "100", "--dry-run"],
        )
        with pytest.raises(SystemExit) as exc_info:
            main()

        assert exc_info.value.code == 1
        captured = capsys.readouterr()
        assert "not found" in captured.err
